# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module implements experimental propagators to inject trace context
into response carriers. This is useful for server side frameworks that start traces
when server requests and want to share the trace context with the client so the
client can add its spans to the same trace.

This is part of an upcoming W3C spec and will eventually make it to the Otel spec.

https://w3c.github.io/trace-context/#trace-context-http-response-headers-format
"""

import typing
from abc import ABC, abstractmethod

from opentelemetry import trace
from opentelemetry.context.context import Context
from opentelemetry.propagators import textmap
from opentelemetry.trace import format_span_id, format_trace_id

_HTTP_HEADER_ACCESS_CONTROL_EXPOSE_HEADERS = "Access-Control-Expose-Headers"
_RESPONSE_PROPAGATOR = None


def get_global_response_propagator():
    return _RESPONSE_PROPAGATOR


def set_global_response_propagator(propagator):
    global _RESPONSE_PROPAGATOR  # pylint:disable=global-statement
    _RESPONSE_PROPAGATOR = propagator


class Setter(ABC):
    @abstractmethod
    def set(self, carrier, key, value):
        """Inject the provided key value pair in carrier."""


class DictHeaderSetter(Setter):
    def set(self, carrier, key, value):  # pylint: disable=no-self-use
        old_value = carrier.get(key, "")
        if old_value:
            value = f"{old_value}, {value}"
        carrier[key] = value


class FuncSetter(Setter):
    """FuncSetter converts a function into a valid Setter. Any function that
    can set values in a carrier can be converted into a Setter by using
    FuncSetter. This is useful when injecting trace context into non-dict
    objects such HTTP Response objects for different framework.

    For example, it can be used to create a setter for Falcon response object
    as:

        setter = FuncSetter(falcon.api.Response.append_header)

    and then used with the propagator as:

        propagator.inject(falcon_response, setter=setter)

    This would essentially make the propagator call `falcon_response.append_header(key, value)`
    """

    def __init__(self, func):
        self._func = func

    def set(self, carrier, key, value):
        self._func(carrier, key, value)


default_setter = DictHeaderSetter()


class ResponsePropagator(ABC):
    @abstractmethod
    def inject(
        self,
        carrier: textmap.CarrierT,
        context: typing.Optional[Context] = None,
        setter: textmap.Setter = default_setter,
    ) -> None:
        """Injects SpanContext into the HTTP response carrier."""


class TraceResponsePropagator(ResponsePropagator):
    """Experimental propagator that injects tracecontext into HTTP responses."""

    def inject(
        self,
        carrier: textmap.CarrierT,
        context: typing.Optional[Context] = None,
        setter: textmap.Setter = default_setter,
    ) -> None:
        """Injects SpanContext into the HTTP response carrier."""
        span = trace.get_current_span(context)
        span_context = span.get_span_context()
        if span_context == trace.INVALID_SPAN_CONTEXT:
            return

        header_name = "traceresponse"
        setter.set(
            carrier,
            header_name,
            f"00-{format_trace_id(span_context.trace_id)}-{format_span_id(span_context.span_id)}-{span_context.trace_flags:02x}",
        )
        setter.set(
            carrier,
            _HTTP_HEADER_ACCESS_CONTROL_EXPOSE_HEADERS,
            header_name,
        )



import greenlet
from . import TestCase


class Test(TestCase):

    def test_stack_saved(self):
        main = greenlet.getcurrent()
        self.assertEqual(main._stack_saved, 0)

        def func():
            main.switch(main._stack_saved)

        g = greenlet.greenlet(func)
        x = g.switch()
        self.assertGreater(x, 0)
        self.assertGreater(g._stack_saved, 0)
        g.switch()
        self.assertEqual(g._stack_saved, 0)



import json
import re
import xml.sax.handler

from cache_functions import UsersInfoCache
from typing import (
    Dict,
    Optional,
)

from production.model import ProductionOrder, Item, Comment, StateEvent


def xml_str(v):
    return str(v) if v is not None else None


def xml_bool(v):
    return "true" if (v and str(v).lower() == "true") else "false"


def json_object(value):
    try:
        return json.loads(value)
    except (Exception,):
        return None


class SystemOrderParser(xml.sax.handler.ContentHandler):
    # [production prop], [<Order> attr], [type converter]
    # <Order> attributes to be merged
    _ORDER_MERGE = (
        ('state', 'state', str),
        ('state_id', 'stateId', int),
        ('originator', 'originatorId', str),
        ('originator_number', 'originatorNumber', int),
        ('pos_id', 'posId', int),
        ('pod_type', 'podType', str),
        ('order_type', 'type', str),
        ('order_subtype', 'subType', str),
        ('sale_type', 'saleTypeDescr', str),
        ('total_gross', 'totalGross', str),
        ('tax_total', 'taxTotal', str),
    )
    # ALL <Order> attributes
    _ORDER_ATTRS = _ORDER_MERGE + (
        ('order_id', 'orderId', str),
        ('period', 'businessPeriod', int),
        ('pod_type', 'podType', str),
        ('created_at', 'createdAt', str),
        ('created_at_gmt', 'createdAtGMT', str),
        ('business_day', 'businessPeriod', str),
        ('operator_id', 'operatorId', int),
        ('session_id', 'sessionId', str),
        ('selected_pickup_area', 'pickupArea', str),
    )
    # <SaleLine> attributes to be merged
    _LINE_MERGE = (
        ('qty', 'qty', float),
        ('added_qty', 'addedQty', float),
        ('sub_qty', 'subQty', float),
        ('multiplied_qty', 'multipliedQty', float),
        ('modifier_label', 'modifierLabel', xml_str),
        ('only', 'onlyModifiers', xml_bool),
        ('properties', 'customProperties', json_object),
    )
    # noinspection PyTypeChecker
    # ALL <SaleLine> attributes
    _LINE_ATTRS = _LINE_MERGE + (
        ('line_number', 'lineNumber', int),
        ('item_id', 'itemId', str),
        ('part_code', 'partCode', int),
        ('description', 'productName', unicode),  # noqa: F821
        ('item_type', 'itemType', str),
        ('level', 'level', float),
        ('pos_level', 'level', float),
        ('default_qty', 'defaultQty', float),
        ('min_qty', 'minQty', float),
        ('max_qty', 'maxQty', float),
        ('last_order_qty', 'lastOrderedQty', float),
        ('product_priority', 'productPriority', int),
        ('json_tags', 'jsonArrayTags', str),
        ('last_updated', 'lastUpdated', str),
        ('measure_unit', 'measureUnit', str),
        ('created_at', 'createdAt', str)
    )
    # <-- ALL <State> attributes
    _STATE_ATTRS = (
        ('state', 'state', str),
        ('state_id', 'stateId', int),
        ('timestamp', 'timestamp', str),
    )

    current_item = None

    def __init__(
        self,                           # type: SystemOrderParser
        part_code_jit_lines,            # type: Dict[int, str]
        product_custom_params,          # type: Dict[str, Dict[int, str]]
        merge_existing=True,            # type: Optional[bool]
        user_cache_thread=None,         # type: Optional[UsersInfoCache]
    ):
        # type: (...) -> None

        self.part_code_jit_lines = part_code_jit_lines
        self.product_custom_params = product_custom_params
        self.merge = merge_existing
        self.user_cache_thread = user_cache_thread

        self.order = None                       #: Contains the resulting order (after the parsing is done)
        self.saved_order = None                 #: Contains the original order (for comparing after parsing)
        self._stack = {}                        #: hierarchical pile of items
        self.current_item = None                #: item being parsed
        self.mainItem = None                    #: level 0 item

        if merge_existing:
            self._missing = set()

    @staticmethod
    def _create_attributes(tuples, attrs):
        d = {}
        for my_name, other_name, func in tuples:
            value = attrs.get(other_name)
            if value or hasattr(func, 'allowempty'):
                d[my_name] = func(value)
        return d

    def startElement(self, name, attrs):
        if name == "Order":
            self._parse_order(attrs)
        elif name == "SaleLine":
            self._parse_line(attrs)
        elif name == "State":
            self._parse_state(attrs)
        elif name == "OrderProperty":
            self._parse_property(attrs)
        elif name == "Comment":
            self._parse_comment(attrs)

    def endElement(self, name):
        if self.merge and name == "Order" and self._missing:
            # After all the order has been merged, void items that where not found
            for item in self._missing:
                if item.qty > 0:
                    item.qty = 0
        if self.saved_order is not None:
            # Check if the order has changed
            self.order.modified = self.order.equal_items(self.saved_order) is False

        # Check voided items
        for item in self.order.items:
            if item.qty == 0:
                self._mark_voided(item)
            elif item.voided is True:
                self._clear_voided(item)

    def endDocument(self):
        self._merge_fraction_lines()

    def _mark_voided(self, item):
        item.voided = True
        for sub_item in item.items:
            self._mark_voided(sub_item)

    def _clear_voided(self, item):
        item.voided = False
        for sub_item in item.items:
            self._clear_voided(sub_item)

    def _parse_order(self, attrs):
        if self.order is None:  # Create a new order
            kargs = self._create_attributes(self._ORDER_ATTRS, attrs)
            self.order = ProductionOrder(**kargs)
            self._update_operator_info()
        else:  # Merge attributes
            self.saved_order = self.order.clone()
            order = self.order
            for key, value in self._create_attributes(self._ORDER_MERGE, attrs).items():
                setattr(order, key, value)
            # Populate the list of items, so that we can find "missing ones" later
            self._missing = set(order.items)

    def _merge_fraction_lines(
        self,   # type: SystemOrderParser
    ):
        # type: (...) -> None

        merged_items = []
        for item in self.order.items:
            merged_items.append(item)
            if item.level != 0 or "OMGR_FRACTION_FROM_LINE" not in item.properties:
                continue

            original_line_number = int(item.properties.get("OMGR_FRACTION_FROM_LINE"))
            if item.line_number == original_line_number:
                fractionated_by = int(item.properties.get("fractionatedBy", 1))
                for from_item in merged_items:
                    from_item.multiplied_qty = item.multiplied_qty * fractionated_by

                continue

            for from_item in merged_items:
                if from_item.line_number != original_line_number:
                    continue

                from_item.qty += item.qty
                self._set_item_multiplied_qty(
                    from_item=from_item,
                    original_item=item,
                )

                break

            merged_items.remove(item)

        self.order.items = merged_items

    def _set_item_multiplied_qty(
        self,           # type: SystemOrderParser
        from_item,      # type: Item
        original_item,  # type: Item
    ):
        # type: (...) -> None

        from_item.multiplied_qty += (original_item.multiplied_qty * from_item.qty)
        for child in from_item.items:
            self._set_item_multiplied_qty(
                from_item=child,
                original_item=original_item,
            )

    def _update_operator_info(self):
        self._update_operator_id()

        users = self.user_cache_thread.get_users_info() if self.user_cache_thread else []
        if self.order.operator_id in users:
            user = users[self.order.operator_id]
            self._update_operator_name(user.name)
            self._update_operator_level(user.level)

    def _update_operator_id(self):
        self.order.operator_id = self._parse_user_from_session_id() if self._session_id_has_user() else "0"

    def _update_operator_name(self, name):
        self.order.operator_name = name

    def _update_operator_level(self, level):
        self.order.operator_level = level

    def _parse_user_from_session_id(self):
        return self.order.session_id.split("user=")[1].split(",")[0]

    def _session_id_has_user(self):
        return "user=" in self.order.session_id

    def _parse_line(self, attrs):
        item = self._find_item(attrs) if self.merge else None
        if item is None:
            # Create a new item
            kargs = self._create_attributes(self._LINE_ATTRS, attrs)
            item = Item(**kargs)

            if item.last_updated:
                item.last_updated += "Z"

            item_course = 'default'
            if item.properties is not None:
                if "fractionatedBy" in item.properties:
                    fractionated_by = item.properties["fractionatedBy"]
                    item.fractionated_by = fractionated_by

                if "seat" in item.properties:
                    seat = item.properties["seat"]
                    item.seat = seat if seat != "0" else ''

                if "courseAbbreviation" in item.properties:
                    course = item.properties["courseAbbreviation"]
                    item.manual_course = course

                if "course" in item.properties:
                    item_course = (item.properties["course"]).lower()

                if "fire" in item.properties:
                    fire = item.properties["fire"].lower() == "true"
                    item.fire = fire

                if "highPriority" in item.properties:
                    high_priority = item.properties["highPriority"].lower() == "true"
                    item.high_priority = high_priority

                try:
                    if "sequence" in item.properties:
                        sequence = json.loads(item.properties["sequence"])
                        item.sequence = int(sequence["kitchenSequence"]) if "kitchenSequence" in sequence else 1000
                except (TypeError, ValueError):
                    item.sequence = 1000

                if "fraction" in item.properties:
                    item.fraction = int(item.properties["fraction"])

            jit_lines = self.part_code_jit_lines.get(item.part_code, "")
            item.jit_lines = self.get_part_code_jit_by_course(jit_lines.replace(" ", ""), item_course)

            for custom_param_id in self.product_custom_params.keys():
                if item.part_code in self.product_custom_params[custom_param_id]:
                    item.custom_params = {
                        custom_param_id: self.product_custom_params[custom_param_id].get(item.part_code)
                    }

            # Put it on its parent list
            level = float(item.level)
            parent = self._get_parent(level)
            if parent:
                parent.items.append(item)
        else:  # Just merge the attributes
            for key, value in self._create_attributes(self._LINE_MERGE, attrs).items():
                setattr(item, key, value)

        if item.level == 0:
            self.mainItem = item
        elif item.level > 0:
            item.main_item_description = self.mainItem.description

        # Record where we are
        self._stack[str(item.level)] = item
        self.current_item = item

    @staticmethod
    def get_part_code_jit_by_course(jit_lines, item_course):
        part_code_jit = jit_lines.strip()
        part_code_jit_list = part_code_jit.split(';')
        jit_dict = {}

        for jit_item in part_code_jit_list:
            try:
                course = re.findall('{(.*)}', jit_item)[0]
            except IndexError:
                course = "default"

            jit_dict[course.lower()] = jit_item.replace("{" + course + "}", "")

        return jit_dict[item_course] if item_course in jit_dict else jit_dict['default']

    def _parse_comment(self, attrs):
        description = attrs.get("comment")
        if description is not None:
            comment = Comment()
            comment.comment_id = attrs.get("commentId")
            comment.text = description
            self.current_item.comments[comment.comment_id] = comment

    def _parse_property(self, attrs):
        ignore_keys = [
            "FISCAL_XML",
            "CANCELED_FISCAL_XML",
            "DISABLED_FISCAL_XML",
            "FISCAL_ID",
            "FISCALIZATION_DATE",
        ]

        key = attrs.get('key')
        if key in ignore_keys:
            return
        value = attrs.get('value')
        if value and self.order:
            self.order.properties[key] = value

    def _find_item(self, attrs):
        level = float(attrs.get("level"))
        line = int(attrs.get("lineNumber"))
        item_code = "%s.%s" % (attrs.get("itemId"), attrs.get("partCode"))

        parent = self._get_parent(level)
        if not parent:
            return None
        for item in parent.items:
            if (item.line_number == line) and (item.item_code == item_code) and (item.level == level):
                self._missing.discard(item)
                return item
        return None  # Not found

    def _get_parent(self, level):
        if level == 0:
            return self.order
        return self._stack.get(str(level - 1), None) or self._stack.get(str(float(int(level) - 1)), None)

    def _parse_state(self, attrs):
        kargs = self._create_attributes(self._STATE_ATTRS, attrs)
        new = StateEvent(**kargs)
        for event in self.order.state_history:
            if event.state_id == new.state_id and event.timestamp == new.timestamp:
                break  # This event is already in the order
        else:
            # Event not found... add it!
            self.order.state_history.append(new)



# ruc.py - functions for handling Paraguay RUC numbers
# coding: utf-8
#
# Copyright (C) 2019 Leandro Regueiro
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""RUC number (Registro Único de Contribuyentes, Paraguay tax number).

The Registro Único del Contribuyente (RUC) is the unique taxpayer registry
that maintains identification numbers for all persons (national or foreign)
and legal entities in Paraguay.

The RUC number for legal entities consists of 8 digits starting after
80000000. Number for residents and foreigners are up to 9 digits. The last
digit is a check digit.

More information:

* https://www.ruc.com.py/

>>> validate('80028061-0')
'800280610'
>>> validate('9991603')
'9991603'
>>> validate('2660-3')
'26603'
>>> validate('800532492')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> validate('80123456789')
Traceback (most recent call last):
    ...
InvalidLength: ...
>>> format('800000358')
'80000035-8'
"""

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation.

    This strips the number of any valid separators and removes surrounding
    whitespace.
    """
    return clean(number, ' -').upper().strip()


def calc_check_digit(number):
    """Calculate the check digit.

    The number passed should not have the check digit included.
    """
    s = sum((i + 2) * int(n) for i, n in enumerate(reversed(number)))
    return str((-s % 11) % 10)


def validate(number):
    """Check if the number is a valid Paraguay RUC number.

    This checks the length, formatting and check digit.
    """
    number = compact(number)
    if len(number) > 9:
        raise InvalidLength()
    if not isdigits(number):
        raise InvalidFormat()
    if number[-1] != calc_check_digit(number[:-1]):
        raise InvalidChecksum()
    return number


def is_valid(number):
    """Check if the number is a valid Paraguay RUC number."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number):
    """Reformat the number to the standard presentation format."""
    number = compact(number)
    return '-'.join([number[:-1], number[-1]])



import json
from xml.etree import cElementTree as eTree

from bustoken import TK_SLCTRL_OMGR_FRACTIONATELINE
from msgbus import TK_SYS_ACK
from mw_helper import (
    show_keyboard,
    show_message_dialog,
    show_messagebox,
)
from sysactions import (
    StopAction,
    action,
    send_message,
)
from typing import (
    Dict,
    List,
    Optional,
    Union,
)

from custom_sysactions import user_authorized_to_execute_action

INFORMATION_TITLE = "$INFORMATION"


@action
@user_authorized_to_execute_action
def do_split_items(
    pos_id,            # type: str
    sale_line_json,    # type: json
):
    # type: (...) -> Optional[bool]

    from msgbusboundary import MsgBusTableService as tableService

    sale_line_list = filter_sale_line_list(
        pos_id=pos_id,
        sale_line_json=sale_line_json,
    )

    if not sale_line_list:
        return False

    table_id = tableService().get_current_table_id(pos_id=pos_id)
    table_seats = tableService().get_service_seats(
        pos_id=pos_id,
        table_id=table_id,
    )

    fraction_qty = show_keyboard(
        posid=pos_id,
        message="",
        title="$FRACTION_QTY",
        mask="INTEGER",
        numpad=True,
        defvalue=table_seats,
    )

    if fraction_qty is None:
        return False

    is_valid_fraction_qty = validate_fraction_qty(
        pos_id=pos_id,
        fraction_qty=fraction_qty,
        message="$INVALID_FRACTION_QTY",
    )
    if not is_valid_fraction_qty:
        return False

    distribute_items_in_seats = False
    if int(fraction_qty) <= table_seats:
        ret = show_messagebox(
            posid=pos_id,
            message="$DISTRIBUTE_ITEMS_IN_SEATS",
            timeout=720000,
            buttons="$YES|$NO",
        )
        distribute_items_in_seats = ret == 0

    split_items(
        pos_id=pos_id,
        table_id=table_id,
        sale_line_list=sale_line_list,
        fraction_qty=fraction_qty,
        distribute_items_in_seats=distribute_items_in_seats,
    )


def validate_fraction_qty(
    pos_id,         # type: str
    fraction_qty,   # type: int
    message,        # type: str
):
    # type: (...) -> bool

    from table_actions import item_division_limit

    if fraction_qty == "" or int(fraction_qty) <= 1 or int(fraction_qty) > item_division_limit:
        show_message_dialog(
            pos_id=pos_id,
            title=INFORMATION_TITLE,
            message=message,
        )
        return False

    return True


def split_items(
    pos_id,                     # type: str
    table_id,                   # type: str
    sale_line_list,             # type: List
    fraction_qty,               # type: int
    distribute_items_in_seats,  # type: bool
):
    # type: (...) -> Optional[bool]

    from msgbusboundary import MsgBusTableService as tableService
    from table_actions import create_table_modified_event
    from api.table import get_table_service

    orders_data = _get_orders_data(sale_line_list=sale_line_list)
    table_service = get_table_service(pos_id=pos_id)
    table = table_service.get_table(table_id=table_id)

    data = {
        "orders_data": orders_data,
        "blk_notify": True,
        "pos_id": int(pos_id),
        "fraction_qty": int(fraction_qty),
        "occupied_seats": table.occupied_seats if distribute_items_in_seats else [],
    }

    msg = send_message(
        dest_name="ORDERMGR{}".format(pos_id),
        token=TK_SLCTRL_OMGR_FRACTIONATELINE,
        data=json.dumps(data),
        timeout=60 * 1000000,
    )
    if msg.token != TK_SYS_ACK:
        show_message_dialog(pos_id, "$FATAL_ERROR_CALL_SUPPORT", msg.data)
        raise StopAction()

    table_picture = tableService().get_table_picture(
        pos_id=pos_id,
        table_id=table_id,
    )
    create_table_modified_event(
        pos_id=pos_id,
        table_picture=table_picture,
    )
    return True


def _get_orders_data(
    sale_line_list,     # type: List
):
    # type: (...) -> Dict[str, List]

    orders_data = {}
    for sale_line in sale_line_list:
        order_id = sale_line.get("orderId", None)
        line_number = sale_line.get("lineNumber", None)
        if order_id in orders_data:
            orders_data[order_id].append(line_number)
        else:
            orders_data[order_id] = [line_number]

    return orders_data


def filter_sale_line_list(
    pos_id,            # type: str
    sale_line_json,    # type: json
):
    # type: (...) -> Union[List, bool]

    sale_line_list = json.loads(sale_line_json)
    sale_line_list = list(filter(lambda x: x.get("level") == 0, sale_line_list))

    if len(sale_line_list) < 1:
        show_message_dialog(
            pos_id=pos_id,
            title=INFORMATION_TITLE,
            message="$NO_LINE_SELECTED"
        )
        return False

    new_sale_line_list = []
    for sale_line in sale_line_list:
        if sale_line["fractionFrom"] is None:
            new_sale_line_list.append(sale_line)

    if not new_sale_line_list:
        show_message_dialog(
            pos_id=pos_id,
            title=INFORMATION_TITLE,
            message="$CANNOT_FRACTION_TWICE",
        )
        return False

    return new_sale_line_list



class Threshold(object):
    def __init__(self, value):
        # type: (unicode) -> None

        values = value.split("#")
        self.time = values[0]
        self.color = "#" + values[1]
        self.percentage = None
        if len(values) > 2:
            self.percentage = "#" + values[2]



# -*- coding: utf-8 -*-
import logging

import utils
import actions.util

from threading import Lock
from xml.etree import cElementTree as eTree
from typing import (
    Optional,
    List,
)

from mw_helper import show_message_options_dialog

from actions.login.common_lock import release_user_lock
from actions.login.get_lock_message import get_lock_message
from actions.login.get_user_info import get_user_info
from actions.login.lock_user_actions import acquire_user_lock_close_operator
from actions.login.pause_user import pause_user
from actions.util import get_pos_users
from actions.util.list_users import list_all_opened_users
from model.customexception import (
    GettingUserInfoException,
    CanceledDialogException,
)

from sysactions import (
    action,
    get_model,
    show_messagebox,
    StopAction,
    AuthenticationFailed,
    get_poslist,
    close_asynch_dialog,
    force_send_updated_model_to_ui,
)

from login_user import sign_in_quick_user
from signin_user import sign_in_table_user
from actions.pos import pos_must_be_blocked

current_pos_login_users = {}
login_lock = Lock()
mbcontext = None
logger = logging.getLogger("PosActions")


@action
def common_login(
    pos_id,                     # type: str
    user_id=None,               # type: Optional[str]
):
    # type: (...) -> bool

    global current_pos_login_users

    user_id = str(int(user_id)) if user_id is not None else user_id
    model = get_model(pos_id)
    try:
        user_info = get_user_info(pos_id, user_id, model)
        utils.validate_user_level(pos_id, user_info)
    except CanceledDialogException:
        raise StopAction()
    except (AuthenticationFailed, GettingUserInfoException):
        raise StopAction()

    user_id = user_info.get("UserId")
    with login_lock:
        logged_pos_id = current_pos_login_users[user_id] if user_id in current_pos_login_users else None

    if logged_pos_id is not None:
        show_messagebox(
            posid=pos_id,
            message="$USER_IS_LOGIN_IN_POS|{}".format(logged_pos_id),
            icon="error",
        )
        raise StopAction()

    if not acquire_user_lock_close_operator(user_id=user_id):
        show_messagebox(
            posid=pos_id,
            message=get_lock_message(user_id=user_id),
            icon="warning",
        )
        raise StopAction()

    try:
        with login_lock:
            current_pos_login_users[user_id] = pos_id

        can_open_on_blocked_day = _can_open_on_blocked_day(
            pos_id,
            user_id,
        )

        working_mode = model.find("WorkingMode").get("usrCtrlType")

        _login_validations(
            pos_id=pos_id,
            user_id=user_id,
            can_open_on_blocked_day=can_open_on_blocked_day,
            working_mode=working_mode,
        )

        if working_mode == "QSR":
            return sign_in_quick_user(pos_id, user_info)
        else:
            return sign_in_table_user(pos_id, user_info)
    finally:
        with login_lock:
            current_pos_login_users[user_id] = None
            release_user_lock(user_id)


def _can_open_on_blocked_day(
    pos_id,                 # type: str
    user_id,                # type: str
):
    # type: (...) -> bool

    can_open_on_blocked_day = False
    users = get_pos_users(pos_id, [user_id])
    if len(users) > 0:
        current_user = users[0]
        can_open_on_blocked_day = current_user.get("openTime") is not None and current_user.get("closeTime") is None

    return can_open_on_blocked_day


def _login_validations(
    pos_id,                     # type: str
    user_id,                    # type: str
    can_open_on_blocked_day,    # type: bool
    working_mode,               # type: str
):
    if not can_open_on_blocked_day and pos_must_be_blocked(pos_id):
        show_messagebox(pos_id, message="$POS_BLOCKED", icon="error")
        raise StopAction()

    if not utils.check_current_day(pos_id):
        raise StopAction()

    actions.util.set_context(mbcontext)
    if not actions.util.is_user_already_logged(pos_id, user_id):
        return

    if working_mode != "TS":
        raise StopAction()

    if not can_change_user_pos(
        user_id=user_id,
        current_pos_id=pos_id,
    ):
        raise StopAction()


def can_change_user_pos(
    user_id,            # type: str
    current_pos_id,     # type: str
):
    # type: (...) -> bool

    user_info = _get_current_user_info(
        current_pos_id=current_pos_id,
        user_id=user_id,
    )
    if user_info is None:
        return False

    pos_to_pause_user = user_info.get("posId")
    model = get_model(posid=pos_to_pause_user)

    operator_state = _get_operator_state_from_model(
        model=model,
        user_id=user_id,
    )
    _log_message(message="Operator {} is {} in pos {}".format(str(user_id), operator_state, pos_to_pause_user))

    is_table_service = model.find("WorkingMode").get("usrCtrlType") == "TS"
    if not is_table_service:
        return False

    _show_confirm_message_open_pos(
        current_pos_id=current_pos_id,
        pos_to_pause_user=pos_to_pause_user,
    )

    utils.request_operator_confirmation(
        pos_id=current_pos_id,
        pos_to_close_user=pos_to_pause_user,
    )

    return _validate_orders(
        model=model,
        current_pos_id=current_pos_id,
        user_id=user_id,
        pos_to_pause_user=pos_to_pause_user,
    )


def _get_current_user_info(
    current_pos_id,     # type: str
    user_id,            # type: str
):
    # type: (...) -> Optional[eTree.Element]

    for pos_id in sorted(get_poslist()):
        user_info = _get_pos_user_info(
            pos_id_to_list=pos_id,
            current_pos_id=current_pos_id,
            user_id=user_id,
        )

        if user_info is not None:
            return user_info


def _get_pos_user_info(
    pos_id_to_list,     # type: str
    current_pos_id,     # type: str
    user_id,            # type: str
):
    # type: (...) -> Optional[eTree.Element]

    if str(pos_id_to_list) == current_pos_id:
        return None

    opened_users = list_all_opened_users(pos_id=str(pos_id_to_list))
    if len(opened_users) == 0:
        return None

    user_info = _get_user_info(
        current_pos_id=current_pos_id,
        opened_users=opened_users,
        user_id=user_id,
    )

    return user_info


def _get_user_info(
    current_pos_id,     # type: str
    opened_users,       # type: List[eTree.Element]
    user_id,            # type: str
):
    # type: (...) -> Optional[eTree.Element]

    for opened_user in opened_users:
        if _validate_user_parameters(
            current_pos_id=current_pos_id,
            user_id=user_id,
            user=opened_user,
        ):
            return opened_user

    return None


def _validate_user_parameters(
    current_pos_id,     # type: str
    user_id,            # type: str
    user,               # type: eTree.Element
):
    # type: (...) -> bool

    return user.get("id") == user_id and user.get("userControlType") == "TS" and user.get("posId") != current_pos_id


def _get_operator_state_from_model(
    model,      # type: eTree
    user_id,    # type: str
):
    # type: (...) -> Optional[str]

    operator = model.find("Operators/Operator[@id='{user_id}']".format(user_id=user_id))
    return operator.get('state') if operator is not None else None


def _show_confirm_message_open_pos(
    current_pos_id,     # type: str
    pos_to_pause_user,  # type: str
):
    # type: (...) -> None

    response = show_message_options_dialog(
        pos_id=current_pos_id,
        buttons="$OK|$PREVIOUS",
        message="$CONFIRM_OPENING_OPERATOR_LOGGED_IN_ANOTHER_POS|{}".format(pos_to_pause_user),
    )
    if response is None or response == 1:
        raise StopAction()


def _validate_orders(
    model,              # type: eTree.Element
    current_pos_id,     # type: str
    user_id,            # type: str
    pos_to_pause_user,  # type: str
):
    # type: (...) -> bool

    utils.close_current_order(
        model=model,
        current_pos_id=current_pos_id,
        pos_to_close_order=pos_to_pause_user,
        user_id=user_id,
    )

    _log_message(message="Init process pause_user")
    pause_user(
        pos_id=pos_to_pause_user,
        pos_id_to_show_message=current_pos_id,
    )

    close_asynch_dialog(
        posid=pos_to_pause_user,
        dlgid=-1,
    )

    force_send_updated_model_to_ui(pos_to_pause_user)
    _log_message(message="End process pause_user")

    return True


def _log_message(
    message,    # type: str
):
    # type: (...) -> None

    logger.info("[Change Login]: {}".format(message))



from dataclasses import dataclass


@dataclass
class AuthenticationData:

    generate_by: str
    user_id: str
    user_name: str
    long_name: str
    user_level: int
    authentication_key: str
    generation_date: str
    expiration_date: str
    interval: int



from ._FiscalCouponNfceBuilder import FiscalCouponNfceBuilder
from ._FiscalCouponSatBuilder import FiscalCouponSatBuilder
from ._FiscalCouponBuilder import FiscalCouponBuilder
from ._BaseFiscalNfce import BaseFiscalNfce
from ._BaseFiscalSat import BaseFiscalSat



try:
    from inspect import getfullargspec as get_args
except ImportError:
    from inspect import getargspec as get_args

from typing import (
    Any,
    Optional,
    List,
    Dict,
    Callable,
)
from messagebus import MessageBus, Message, Event

from ._MessageProcessor import MessageProcessor
from ._MessageProcessorCallback import MessageProcessorCallback


class DefaultMessageProcessorExecutor(object):
    def __init__(self, event_processor, callbacks=None):
        # type: (MessageProcessor, Optional[List[MessageProcessorCallback]]) -> None
        self.event_processor = event_processor
        self.callbacks = callbacks or []

    def execute_processor(self, message_bus, message, event, data):
        # type: (MessageBus, Optional[Message], Optional[Event], Any) -> None
        event_name = self.event_processor.get_processor_name()
        request_id = self.event_processor.generate_unique_id()

        base_params = dict(
            data=data,
            event_type=event.event_type if event is not None else None,
            event=event,
            message=message,
        )
        self._process_begun(event_name, request_id, **base_params)
        input_obj = None
        data_parsed = False
        try:
            kwargs = self._get_function_necessary_parameters(
                func=self.event_processor.parse_data,
                **base_params
            )

            input_obj = self.event_processor.parse_data(**kwargs)
            data_parsed = True

            self._input_parsed(event_name, request_id, input_obj)
            event_name = self._reevaluate_event_name_after_input_is_parsed()

            output_obj = self.event_processor.call_business(input_obj)
            self._business_called(event_name, request_id, input_obj, output_obj)
            self.event_processor.format_response(message_bus, message, event, input_obj, output_obj)
            self._process_finished(event_name, request_id, input_obj, output_obj)
        except BaseException as ex: # noqa
            if data_parsed:
                self.event_processor.format_exception(message_bus, message, event, input_obj, ex)
                self._unhandled_exception(event_name, request_id, input_obj, ex)
            else:
                self.event_processor.format_parse_exception(message_bus, message, event, data, ex)
                self._parse_exception(event_name, request_id, data, ex)

    def _get_function_necessary_parameters(
        self,
        data,               # type: bytes
        event,              # type: Optional[Event]
        event_type,         # type: Optional[str]
        message,            # type: Optional[Message]
        func,               # type: Callable
    ):
        # type: (...) -> Dict[str, Any]

        full_kw_args = {
            "data": data,
            "event_type": event_type,
            "event": event,
            "message": message,
        }

        args_spec = get_args(func)

        if hasattr(args_spec, "varkw") and args_spec.varkw is not None:
            return full_kw_args
        if hasattr(args_spec, "varargs") and args_spec.varargs is not None:
            return full_kw_args

        kwargs = {}
        keys = ("data", "event", "event_type", "message")
        for key in keys:
            if key in args_spec.args:
                kwargs[key] = full_kw_args[key]
        return kwargs

    def _reevaluate_event_name_after_input_is_parsed(self,):
        return self.event_processor.get_processor_name()

    def _process_begun(self, event_name, request_id, **kwargs):
        # type: (str, str, Dict[str, Any]) -> None
        for callback in self.callbacks:
            func_kwargs = self._get_function_necessary_parameters(
                func=callback.process_begun,
                **kwargs
            )
            callback.process_begun(event_name, request_id, **func_kwargs)

    def _input_parsed(self, event_name, request_id, input_obj):
        # type: (str, str, Any) -> None
        for callback in self.callbacks:
            callback.input_parsed(event_name, request_id, input_obj)

    def _business_called(self, event_name, request_id, input_obj, output_obj):
        # type: (str, str, Any, Any) -> None
        for callback in self.callbacks:
            callback.business_called(event_name, request_id, input_obj, output_obj)

    def _process_finished(self, event_name, request_id, input_obj, output_obj):
        # type: (str, str, Any, Any) -> None
        for callback in self.callbacks:
            callback.process_finished(event_name, request_id, input_obj, output_obj)

    def _parse_exception(self, event_name, request_id, data, exception):
        # type: (str, str, bytes, BaseException) -> None
        for callback in self.callbacks:
            callback.parse_exception(event_name, request_id, data, exception)

    def _unhandled_exception(self, event_name, request_id, input_obj, exception):
        # type: (str, str, Any, BaseException) -> None
        for callback in self.callbacks:
            callback.unhandled_exception(event_name, request_id, input_obj, exception)



class CodedError(Exception):
    def __init__(
        self,
        code,           # type: int
        message,        # type: str
    ):
        # type: (...) -> None

        self.code = code
        self.message = message


class UpdateItemsError(object):
    INVALID_JSON = CodedError(
        code=-1,
        message="data must be a valid json",
    )

    NO_USER_ID = CodedError(
        code=-2,
        message="input json must contain a user_id field",
    )

    NO_ENABLED = CodedError(
        code=-3,
        message="input json must contain an enabled field",
    )

    NO_DISABLED = CodedError(
        code=-4,
        message="input json must contain a disabled field",
    )

    ENABLED_NOT_A_LIST = CodedError(
        code=-5,
        message="enabled field must be a list of integer part codes",
    )

    DISABLED_NOT_A_LIST = CodedError(
        code=-6,
        message="disabled field must be a list of integer part codes",
    )

    ENABLED_NOT_A_LIST_OF_INT = CodedError(
        code=-7,
        message="enabled field must be a list of integer part codes",
    )

    DISABLED_NOT_A_LIST_OF_INT = CodedError(
        code=-8,
        message="enabled field must be a list of integer part codes",
    )

    INVALID_PART_CODE = CodedError(
        code=-9,
        message="",
    )


class RuptureItemsByQuantityError(object):

    INVALID_NUMBER_OF_PARAMS = CodedError(
        code=-1,
        message="Invalid numbers of parameters",
    )

    INVALID_JSON = CodedError(
        code=-2,
        message="Data must be a valid JSON",
    )

    NO_USER_ID = CodedError(
        code=-3,
        message="User ID must not be empty",
    )

    NOT_INTEGER_USER_ID = CodedError(
        code=-4,
        message="User ID must be integer",
    )

    UPDATE_REQUEST_ITEMS_NOT_A_DICT = CodedError(
        code=-5,
        message="Items details must be a dict",
    )

    SET_AND_UNSET_DOES_NOT_EXISTS = CodedError(
        code=-6,
        message="One of set or unset must exists",
    )

    SET_NOT_A_LIST = CodedError(
        code=-7,
        message="Add to rupture by quantity field must be a list of dict",
    )

    UNSET_NOT_A_LIST = CodedError(
        code=-8,
        message="Remove to rupture by quantity field must be a list of dict",
    )

    SET_ITEMS_NOT_DICTS = CodedError(
        code=-9,
        message="Items of set are not all dicts",
    )

    SET_WITHOUT_PRODUCT_CODE = CodedError(
        code=-10,
        message="product_code must be in set items",
    )

    SET_WITHOUT_PRODUCT_QUANTITY = CodedError(
        code=-11,
        message="product_quantity must be in set items",
    )

    UNSET_ITEMS_NOT_DICTS = CodedError(
        code=-12,
        message="Items of unset are not all dicts",
    )

    UNSET_WITHOUT_PRODUCT_CODE = CodedError(
        code=-13,
        message="product_code must be in set items",
    )

    UNSET_WITHOUT_PRODUCT_QUANTITY = CodedError(
        code=-14,
        message="product_quantity must be in set items",
    )



from logging import Logger
from typing import Dict

from production.box import NonValidJitAddTagBox
from production.model import (
    TagNotFoundException,
    JitNotFoundException,
)
from production.treebuilder import BoxConfiguration

from ._BoxBuilder import BoxBuilder


class NonValidJitAddTagBoxBuilder(BoxBuilder):
    def __init__(
        self,           # type: NonValidJitAddTagBoxBuilder
        loggers,        # type: Dict[str, Logger]
    ):
        # type: (...) -> None

        super(NonValidJitAddTagBoxBuilder, self).__init__(loggers)

    def build(
        self,       # type: NonValidJitAddTagBoxBuilder
        config,     # type: BoxConfiguration
    ):
        # type: (...) -> NonValidJitAddTagBox

        if u"Tag" not in config.extra_config:
            raise TagNotFoundException("NonValidJitAddTagBoxBuilder without tag: {}".format(config.name))

        if u"ValidJitLines" not in config.extra_config:
            raise JitNotFoundException("NonValidJitAddTagBoxBuilder without ValidJitLines: {}".format(config.name))

        return NonValidJitAddTagBox(
            name=config.name,
            sons=config.sons,
            tag=self.get_extra(config, "Tag", None),
            valid_jits=self.get_extra(config, "ValidJitLines", []),
            logger=self.get_logger(config),
        )



# bis.py - functions for handling Belgian BIS numbers
# coding: utf-8
#
# Copyright (C) 2023 Jeff Horemans
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""BIS (Belgian BIS number).


The BIS number (BIS-nummer, Numéro BIS) is a unique identification number
for individuals who are not registered in the National Register, but who
still have a relationship with the Belgian government.
This includes frontier workers, persons who own property in Belgium,
persons with Belgian social security rights but who do not reside in Belgium, etc.

The number is issued by the Belgian Crossroads Bank for Social Security (Banque
Carrefour de la sécurité sociale, Kruispuntbank voor Sociale Zekerheid) and is
constructed much in the same way as the Belgian National Number, i.e. consisting of
11 digits, encoding the person's date of birth and gender, a checksum, etc.
Other than with the national number though, the month of birth of the BIS number
is increased by 20 or 40, depending on whether the sex of the person was known
at the time or not.


More information:

* https://sma-help.bosa.belgium.be/en/faq/where-can-i-find-my-bis-number#7326
* https://www.socialsecurity.be/site_nl/employer/applics/belgianidpro/index.htm
* https://nl.wikipedia.org/wiki/Rijksregisternummer
* https://fr.wikipedia.org/wiki/Numéro_de_registre_national

>>> compact('98.47.28-997.65')
'98472899765'
>>> validate('98 47 28 997 65')
'98472899765'
>>> validate('01 49 07 001 85')
'01490700185'
>>> validate('12345678901')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> format('98472899765')
'98.47.28-997.65'
>>> get_birth_date('98.47.28-997.65')
datetime.date(1998, 7, 28)
>>> get_birth_year('98.47.28-997.65')
1998
>>> get_birth_month('98.47.28-997.65')
7
>>> get_gender('98.47.28-997.65')
'M'
"""

from stdnum.be import nn
from stdnum.exceptions import *
from stdnum.util import isdigits


def compact(number):
    """Convert the number to the minimal representation. This strips the number
    of any valid separators and removes surrounding whitespace."""
    return nn.compact(number)


def validate(number):
    """Check if the number is a valid BIS Number."""
    number = compact(number)
    if not isdigits(number) or int(number) <= 0:
        raise InvalidFormat()
    if len(number) != 11:
        raise InvalidLength()
    nn._get_birth_date_parts(number)
    if not 20 <= int(number[2:4]) <= 32 and not 40 <= int(number[2:4]) <= 52:
        raise InvalidComponent('Month must be in 20..32 or 40..52 range')
    return number


def is_valid(number):
    """Check if the number is a valid BIS Number."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number):
    """Reformat the number to the standard presentation format."""
    return nn.format(number)


def get_birth_year(number):
    """Return the year of the birth date."""
    return nn.get_birth_year(number)


def get_birth_month(number):
    """Return the month of the birth date."""
    return nn.get_birth_month(number)


def get_birth_date(number):
    """Return the date of birth."""
    return nn.get_birth_date(number)


def get_gender(number):
    """Get the person's gender ('M' or 'F'), which for BIS
    numbers is only known if the month was incremented with 40."""
    number = compact(number)
    if int(number[2:4]) >= 40:
        return nn.get_gender(number)



# testing/fixtures/base.py
# Copyright (C) 2005-2025 the SQLAlchemy authors and contributors
# <see AUTHORS file>
#
# This module is part of SQLAlchemy and is released under
# the MIT License: https://www.opensource.org/licenses/mit-license.php
# mypy: ignore-errors


from __future__ import annotations

import sqlalchemy as sa
from .. import assertions
from .. import config
from ..assertions import eq_
from ..util import drop_all_tables_from_metadata
from ... import Column
from ... import func
from ... import Integer
from ... import select
from ... import Table
from ...orm import DeclarativeBase
from ...orm import MappedAsDataclass
from ...orm import registry


@config.mark_base_test_class()
class TestBase:
    # A sequence of requirement names matching testing.requires decorators
    __requires__ = ()

    # A sequence of dialect names to exclude from the test class.
    __unsupported_on__ = ()

    # If present, test class is only runnable for the *single* specified
    # dialect.  If you need multiple, use __unsupported_on__ and invert.
    __only_on__ = None

    # A sequence of no-arg callables. If any are True, the entire testcase is
    # skipped.
    __skip_if__ = None

    # if True, the testing reaper will not attempt to touch connection
    # state after a test is completed and before the outer teardown
    # starts
    __leave_connections_for_teardown__ = False

    def assert_(self, val, msg=None):
        assert val, msg

    @config.fixture()
    def nocache(self):
        _cache = config.db._compiled_cache
        config.db._compiled_cache = None
        yield
        config.db._compiled_cache = _cache

    @config.fixture()
    def connection_no_trans(self):
        eng = getattr(self, "bind", None) or config.db

        with eng.connect() as conn:
            yield conn

    @config.fixture()
    def connection(self):
        global _connection_fixture_connection

        eng = getattr(self, "bind", None) or config.db

        conn = eng.connect()
        trans = conn.begin()

        _connection_fixture_connection = conn
        yield conn

        _connection_fixture_connection = None

        if trans.is_active:
            trans.rollback()
        # trans would not be active here if the test is using
        # the legacy @provide_metadata decorator still, as it will
        # run a close all connections.
        conn.close()

    @config.fixture()
    def close_result_when_finished(self):
        to_close = []
        to_consume = []

        def go(result, consume=False):
            to_close.append(result)
            if consume:
                to_consume.append(result)

        yield go
        for r in to_consume:
            try:
                r.all()
            except:
                pass
        for r in to_close:
            try:
                r.close()
            except:
                pass

    @config.fixture()
    def registry(self, metadata):
        reg = registry(
            metadata=metadata,
            type_annotation_map={
                str: sa.String().with_variant(
                    sa.String(50), "mysql", "mariadb", "oracle"
                )
            },
        )
        yield reg
        reg.dispose()

    @config.fixture
    def decl_base(self, metadata):
        _md = metadata

        class Base(DeclarativeBase):
            metadata = _md
            type_annotation_map = {
                str: sa.String().with_variant(
                    sa.String(50), "mysql", "mariadb", "oracle"
                )
            }

        yield Base
        Base.registry.dispose()

    @config.fixture
    def dc_decl_base(self, metadata):
        _md = metadata

        class Base(MappedAsDataclass, DeclarativeBase):
            metadata = _md
            type_annotation_map = {
                str: sa.String().with_variant(
                    sa.String(50), "mysql", "mariadb"
                )
            }

        yield Base
        Base.registry.dispose()

    @config.fixture()
    def future_connection(self, future_engine, connection):
        # integrate the future_engine and connection fixtures so
        # that users of the "connection" fixture will get at the
        # "future" connection
        yield connection

    @config.fixture()
    def future_engine(self):
        yield

    @config.fixture()
    def testing_engine(self):
        from .. import engines

        def gen_testing_engine(
            url=None,
            options=None,
            future=None,
            asyncio=False,
            transfer_staticpool=False,
            share_pool=False,
        ):
            if options is None:
                options = {}
            options["scope"] = "fixture"
            return engines.testing_engine(
                url=url,
                options=options,
                asyncio=asyncio,
                transfer_staticpool=transfer_staticpool,
                share_pool=share_pool,
            )

        yield gen_testing_engine

        engines.testing_reaper._drop_testing_engines("fixture")

    @config.fixture()
    def async_testing_engine(self, testing_engine):
        def go(**kw):
            kw["asyncio"] = True
            return testing_engine(**kw)

        return go

    @config.fixture()
    def metadata(self, request):
        """Provide bound MetaData for a single test, dropping afterwards."""

        from ...sql import schema

        metadata = schema.MetaData()
        request.instance.metadata = metadata
        yield metadata
        del request.instance.metadata

        if (
            _connection_fixture_connection
            and _connection_fixture_connection.in_transaction()
        ):
            trans = _connection_fixture_connection.get_transaction()
            trans.rollback()
            with _connection_fixture_connection.begin():
                drop_all_tables_from_metadata(
                    metadata, _connection_fixture_connection
                )
        else:
            drop_all_tables_from_metadata(metadata, config.db)

    @config.fixture(
        params=[
            (rollback, second_operation, begin_nested)
            for rollback in (True, False)
            for second_operation in ("none", "execute", "begin")
            for begin_nested in (
                True,
                False,
            )
        ]
    )
    def trans_ctx_manager_fixture(self, request, metadata):
        rollback, second_operation, begin_nested = request.param

        t = Table("test", metadata, Column("data", Integer))
        eng = getattr(self, "bind", None) or config.db

        t.create(eng)

        def run_test(subject, trans_on_subject, execute_on_subject):
            with subject.begin() as trans:
                if begin_nested:
                    if not config.requirements.savepoints.enabled:
                        config.skip_test("savepoints not enabled")
                    if execute_on_subject:
                        nested_trans = subject.begin_nested()
                    else:
                        nested_trans = trans.begin_nested()

                    with nested_trans:
                        if execute_on_subject:
                            subject.execute(t.insert(), {"data": 10})
                        else:
                            trans.execute(t.insert(), {"data": 10})

                        # for nested trans, we always commit/rollback on the
                        # "nested trans" object itself.
                        # only Session(future=False) will affect savepoint
                        # transaction for session.commit/rollback

                        if rollback:
                            nested_trans.rollback()
                        else:
                            nested_trans.commit()

                        if second_operation != "none":
                            with assertions.expect_raises_message(
                                sa.exc.InvalidRequestError,
                                "Can't operate on closed transaction "
                                "inside context "
                                "manager.  Please complete the context "
                                "manager "
                                "before emitting further commands.",
                            ):
                                if second_operation == "execute":
                                    if execute_on_subject:
                                        subject.execute(
                                            t.insert(), {"data": 12}
                                        )
                                    else:
                                        trans.execute(t.insert(), {"data": 12})
                                elif second_operation == "begin":
                                    if execute_on_subject:
                                        subject.begin_nested()
                                    else:
                                        trans.begin_nested()

                    # outside the nested trans block, but still inside the
                    # transaction block, we can run SQL, and it will be
                    # committed
                    if execute_on_subject:
                        subject.execute(t.insert(), {"data": 14})
                    else:
                        trans.execute(t.insert(), {"data": 14})

                else:
                    if execute_on_subject:
                        subject.execute(t.insert(), {"data": 10})
                    else:
                        trans.execute(t.insert(), {"data": 10})

                    if trans_on_subject:
                        if rollback:
                            subject.rollback()
                        else:
                            subject.commit()
                    else:
                        if rollback:
                            trans.rollback()
                        else:
                            trans.commit()

                    if second_operation != "none":
                        with assertions.expect_raises_message(
                            sa.exc.InvalidRequestError,
                            "Can't operate on closed transaction inside "
                            "context "
                            "manager.  Please complete the context manager "
                            "before emitting further commands.",
                        ):
                            if second_operation == "execute":
                                if execute_on_subject:
                                    subject.execute(t.insert(), {"data": 12})
                                else:
                                    trans.execute(t.insert(), {"data": 12})
                            elif second_operation == "begin":
                                if hasattr(trans, "begin"):
                                    trans.begin()
                                else:
                                    subject.begin()
                            elif second_operation == "begin_nested":
                                if execute_on_subject:
                                    subject.begin_nested()
                                else:
                                    trans.begin_nested()

            expected_committed = 0
            if begin_nested:
                # begin_nested variant, we inserted a row after the nested
                # block
                expected_committed += 1
            if not rollback:
                # not rollback variant, our row inserted in the target
                # block itself would be committed
                expected_committed += 1

            if execute_on_subject:
                eq_(
                    subject.scalar(select(func.count()).select_from(t)),
                    expected_committed,
                )
            else:
                with subject.connect() as conn:
                    eq_(
                        conn.scalar(select(func.count()).select_from(t)),
                        expected_committed,
                    )

        return run_test


_connection_fixture_connection = None


class FutureEngineMixin:
    """alembic's suite still using this"""



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from typing import Final

GEO_CONTINENT_CODE: Final = "geo.continent.code"
"""
Two-letter code representing continent’s name.
"""

GEO_COUNTRY_ISO_CODE: Final = "geo.country.iso_code"
"""
Two-letter ISO Country Code ([ISO 3166-1 alpha2](https://wikipedia.org/wiki/ISO_3166-1#Codes)).
"""

GEO_LOCALITY_NAME: Final = "geo.locality.name"
"""
Locality name. Represents the name of a city, town, village, or similar populated place.
"""

GEO_LOCATION_LAT: Final = "geo.location.lat"
"""
Latitude of the geo location in [WGS84](https://wikipedia.org/wiki/World_Geodetic_System#WGS84).
"""

GEO_LOCATION_LON: Final = "geo.location.lon"
"""
Longitude of the geo location in [WGS84](https://wikipedia.org/wiki/World_Geodetic_System#WGS84).
"""

GEO_POSTAL_CODE: Final = "geo.postal_code"
"""
Postal code associated with the location. Values appropriate for this field may also be known as a postcode or ZIP code and will vary widely from country to country.
"""

GEO_REGION_ISO_CODE: Final = "geo.region.iso_code"
"""
Region ISO code ([ISO 3166-2](https://wikipedia.org/wiki/ISO_3166-2)).
"""


class GeoContinentCodeValues(Enum):
    AF = "AF"
    """Africa."""
    AN = "AN"
    """Antarctica."""
    AS = "AS"
    """Asia."""
    EU = "EU"
    """Europe."""
    NA = "NA"
    """North America."""
    OC = "OC"
    """Oceania."""
    SA = "SA"
    """South America."""



import json
import os
import shutil
import tarfile
import time
import zipfile
from datetime import datetime

from systools import sys_parsefilepath

from application.model import (
    Payment,
    PosSaleTypeNavigation,
    UpdateData,
    UpdateStatus,
    UpdateStep,
    UpdateType,
)
from application.model.customexception import (
    ErrorCheckingPackageIntegrity,
    ErrorGenesisDatabaseNotFound,
    ErrorRetrievingJsonCatalog,
    ErrorRetrievingPendingUpdate,
    ErrorUpdatingDatabases,
    NoCatalogJsonFoundException,
    NoCatalogUpdateFoundException,
)
from application.util import LoggerUtil
from helper import (
    json_serialize,
    remove_accents,
)
from typing import (
    Dict,
    Optional,
)
from ._CommonService import CommonService

logger = LoggerUtil.get_logger_name(__name__)

GENESIS_PATH = "../../genesis"


class UpdaterService(object):
    def __init__(self, api_repository, pos_repository, common_repository, working_on_update, configs):
        self.api_repository = api_repository
        self.pos_repository = pos_repository
        self.working_on_update = working_on_update
        self.common_repository = common_repository

        update_type = UpdateType.catalog
        self.local_configs = configs.updaters[update_type.name]
        data_path = sys_parsefilepath("{HVDATADIR}")
        self.backup_directory = sys_parsefilepath(configs.backup_directory)
        default_work_path = os.path.join(self.backup_directory, update_type.name)

        self.data_relative_path = "data"
        self.data_path = os.path.join(data_path, GENESIS_PATH, self.data_relative_path)
        self.server_path = os.path.join(self.data_path, "server")
        self.genesis_databases_path = os.path.join(data_path, "../../genesis/data/server/databases")
        self.genesis_campaign_databases_path = os.path.join(data_path, "../../genesis/data/server/databases-campaign")
        self.backup_path = os.path.join(default_work_path, "backups")
        self.downloads_path = os.path.join(default_work_path, "downloads")
        self.payments_json = os.path.join(self.server_path, "bundles/payments.json")
        self.navigation_json = os.path.join(self.server_path, "bundles/posSaleTypeNavigation.json")
        self.catalog_json = os.path.join(self.genesis_campaign_databases_path, "catalog.json")

    def download_campaign_pending_version(
        self,                           # type: UpdaterService
        need_backoffice_json=False,     # type: bool
        force=False,                    # type: bool
    ):
        # type: (...) -> Optional[str]

        self._update_catalog_json(need_backoffice_json, force=force)
        catalog_json = self._get_json_file_databases_campaign()
        self._set_best_id_database_campaign(catalog_json)
        update_data = self.pos_repository.has_catalog_pending_update_to_apply()
        if update_data:
            menu_id = str(self.get_database_id(update_data.update_name))
            catalog_names = self._get_catalog_names(catalog_json)
            return catalog_names[str(menu_id)]

    def _update_catalog_json(
        self,                       # type: UpdaterService
        need_backoffice_json,       # type: bool
        force=False,                # type: bool
    ):
        try:
            try:
                catalog_json = self._download_catalog_json(force=force)
            except NoCatalogJsonFoundException as ex:
                logger.info(ex.message)
                return
            except NoCatalogUpdateFoundException as ex:
                logger.info(ex.message)
                return

            catalog_json_version = str(catalog_json["version"])
            update_info = json_serialize(dict(id=catalog_json["id"], name=catalog_json_version))
            logger.info("New databases update found: {}".format(update_info))
            self.download_databases_campaign(catalog_json)
        except Exception as ex:
            logger.exception("Error get update_catalog_json")
            if need_backoffice_json:
                raise ex

    @staticmethod
    def _get_catalog_names(catalog_json):
        # type: (Dict) -> Dict

        catalog_names = {}
        for x in catalog_json.get("menus"):
            catalog_names[str(x.get("id"))] = x.get("name")

        return catalog_names

    def download_databases_campaign(
        self,           # type: UpdaterService
        catalog_json,   # type: str
        new_line=True,  # type: Optional[bool]
    ):
        # type: (...) -> None

        catalog_json_id = catalog_json.get("id")
        self._create_databases_campaign_backup(str(catalog_json_id))
        try:
            self._remove_directory_databases_campaign()
            for menu in catalog_json.get("menus"):
                menu_link = menu.get("link")
                if self._download_is_valid(menu) and menu_link is not None:
                    menu_id = menu.get("id")
                    self.download_menu_databases(menu_id, menu_link)
            if new_line:
                self._insert_new_update(catalog_json, UpdateType.catalog)
            self.pos_repository.update_step(catalog_json_id, UpdateStep.all, UpdateType.catalog)
            self.pos_repository.update_status(catalog_json_id, UpdateStatus.applied)
        except Exception as ex:
            self._remove_directory_databases_campaign()
            self._restore_backup_campaign(catalog_json_id)
            logger.info("Error download database campaign")
            raise ex

    @staticmethod
    def _download_is_valid(menu):
        # type: (str) -> bool

        end_in = menu.get("endIn")
        if end_in is not None and datetime.strptime(end_in, "%Y-%m-%d").date() < datetime.now().date():
            return False

        return True

    def _remove_directory_databases_campaign(self):
        # type: () -> None

        try:
            original_path = os.path.abspath(os.getcwd())
            logger.info("Initializing Delete directory database campaign...")
            os.chdir(self.genesis_campaign_databases_path)
            for _, sub_dirs, _ in os.walk("."):
                for dirname in sub_dirs:
                    self.delete_directory(dirname)

            os.chdir(original_path)
            logger.info("Delete directory database campaign successfully!")
        except (Exception,):
            logger.info("Error Delete directory database campaign")
            raise

    def _create_databases_campaign_backup(self, catalog_json_id):
        # type: (str) -> None

        logger.info("Creating databases campaign backup...")

        original_path = os.path.abspath(os.getcwd())
        backup_package = os.path.join(self.backup_path, catalog_json_id + "_databases_campaign.zip")
        with zipfile.ZipFile(backup_package, "w") as zip_object:
            os.chdir(self.genesis_campaign_databases_path)
            for dirname, sub_dirs, files in os.walk("."):
                zip_object.write(dirname)
                for filename in files:
                    zip_object.write(os.path.join(dirname, filename))

        os.chdir(original_path)
        self.test_zip_file(backup_package)
        self.pos_repository.update_step(catalog_json_id, UpdateStep.backup, UpdateType.catalog_json)

        logger.info("Databases backup successfully created!")

    def _download_catalog_json(
        self,                       # type: UpdaterService
        force=False,                # type: bool
    ):
        # type: (...) -> str

        catalog_json = self.retrieve_pending_update()
        catalog_json_id = str(catalog_json.get("id", ''))
        catalog_json_version = str(catalog_json.get("version"))
        if catalog_json is None or catalog_json_id == '':
            raise NoCatalogJsonFoundException("No JSON update found")

        last_catalog_json_id_applied = self._get_last_downloaded_version(UpdateType.catalog_json)
        last_catalog_json_id = last_catalog_json_id_applied.get("updateId")
        last_catalog_json_name = last_catalog_json_id_applied.get("updateName")

        if not force and catalog_json_id == last_catalog_json_id and catalog_json_version == last_catalog_json_name:
            exception_message = "Databases already downloaded: {}".format(json_serialize(catalog_json))
            raise NoCatalogUpdateFoundException(exception_message)

        logger.info("JSON update found")
        with self.working_on_update:
            self._verify_genesis_campaign_directory()

            self._insert_new_update(catalog_json, UpdateType.catalog_json)
            self._store_downloaded_catalog_json(catalog_json)
            self.pos_repository.update_step(catalog_json_id, UpdateStep.all, UpdateType.catalog_json)

        return catalog_json

    @staticmethod
    def get_package_name(update_data):
        # type: (UpdateData) -> str

        return str(update_data.update_id) + ".zip"

    @staticmethod
    def test_zip_file(zip_path):
        # type: (str) -> None

        logger.info("Validating zip file...")

        try:
            zip_file = zipfile.ZipFile(zip_path)
            if zip_file.testzip():
                logger.error("Failed to check package integrity. [{}] is corrupted!".format(zip_path))
                raise ErrorCheckingPackageIntegrity()
            logger.info("Zip file successfully validated!")
        except (Exception,):
            logger.error("Failed to check package integrity. [{}] is corrupted!".format(zip_path))
            raise ErrorCheckingPackageIntegrity()

    def retrieve_pending_update(self):
        # type: () -> str

        try:
            return self.api_repository.retrieve_pending_update()
        except Exception as ex:
            raise ErrorRetrievingPendingUpdate(ex)

    def _get_last_downloaded_version(self, type_id=None):
        # type: (Optional[str]) -> Optional[str]

        return self.pos_repository.get_last_downloaded_version(type_id)

    def _insert_new_update(self, latest_server_update, type_id):
        # type: (dict, UpdateType) -> None

        logger.info("Inserting new update on database...")
        update_id = latest_server_update["id"]
        update_name = remove_accents(str(latest_server_update["version"]))
        return self.pos_repository.insert_new_update(update_id, update_name, type_id)

    def _verify_genesis_directory(self):
        if not os.path.exists(self.genesis_databases_path):
            os.makedirs(self.genesis_databases_path)

    def _verify_genesis_campaign_directory(self):
        self.create_dir(self.genesis_campaign_databases_path, False)

    def download_menu_databases(self, menu_id, menu_link):
        # type: (int, str) -> None

        logger.info("Trying to download databases")

        package_name = str(menu_id) + ".zip"
        update_package = os.path.join(self.downloads_path, package_name)
        logger.info("Downloading package {} and saving in {}".format(package_name, update_package))
        self._store_menu_databases(menu_link, update_package)
        self.test_zip_file(update_package)
        self.extract_zip(update_package, os.path.join(self.genesis_campaign_databases_path, str(menu_id)))

        logger.info("Package successfully stored and validated!")

    def _store_downloaded_catalog_json(self, catalog_dict):
        # type: (Dict) -> None

        logger.info("Writing package on file...")

        with open(self.catalog_json, 'w') as f:
            json.dump(catalog_dict, f)

        logger.info("Catalog successfully stored and validated!")

    def _store_menu_databases(self, menu_link, update_package):
        # type: (str, str) -> None
        package = self.api_repository.get_menu_databases(menu_link)

        logger.info("Writing package on file...")
        with open(update_package, "wb") as zip_object:
            for chunk in package.iter_content(chunk_size=128):
                zip_object.write(chunk)
        logger.info("Package successfully written on file!")

    def _restore_backup_campaign(self, catalog_json_id):
        # type: (str) -> None

        backup_package = os.path.join(self.backup_path, "{}_databases_campaign.zip".format(catalog_json_id))
        self.extract_zip(backup_package, os.path.join(self.genesis_campaign_databases_path), False)

    @staticmethod
    def create_dir(path, delete_if_exists=True):
        # type: (str, bool) -> None

        CommonService.create_dir(path, delete_if_exists)

    def copy_directory(self, source_path, destination_path):
        # type: (str, str) -> None

        self.delete_directory(path=destination_path)
        shutil.copytree(source_path, destination_path)

    def copy_all_files(self, source_path, destination_path):
        # type: (str, str) -> None
        self.create_dir(destination_path, delete_if_exists=False)

        src_files = os.listdir(source_path)
        for file_name in src_files:
            full_file_name = os.path.join(source_path, file_name)
            if os.path.isfile(full_file_name):
                shutil.copy(full_file_name, destination_path)

    @staticmethod
    def copy_all_content(source_path, destination_path):
        # type: (str, str) -> None

        CommonService.copy_all_content(source_path, destination_path)

    @staticmethod
    def save_json(
        data,               # type: Dict
        destination_path    # type: str
    ):
        # type: (str, str) -> None

        with open(destination_path, "wb") as f:
            json.dump(data, f)

    @staticmethod
    def delete_file(
        path    # type: str
    ):
        if os.path.exists(path):
            os.remove(path)

    @staticmethod
    def copy_file(
        original_file_path,     # type: str
        destination_file_path,  # type: str
    ):
        if os.path.exists(original_file_path):
            shutil.copy(original_file_path, destination_file_path)

    @staticmethod
    def move_directory(source_path, destination_path):
        # type: (str, str) -> None

        shutil.move(source_path, destination_path)

    @staticmethod
    def save_directory_zip(source_path, destination_path):
        # type: (str, str) -> None

        UpdaterService._fix_file_timestamps(directory=source_path)
        shutil.make_archive(destination_path, "zip", source_path)

    @staticmethod
    def _fix_file_timestamps(
        directory,  # type: str
    ):
        # type: (...) -> None

        current_time = time.time()
        min_valid_timestamp = UpdaterService._get_min_timestamp()
        for root, dirs, files in os.walk(directory):
            for f in files:
                file_path = os.path.join(root, f)
                try:
                    file_stat = os.stat(file_path)
                    if file_stat.st_mtime < min_valid_timestamp:
                        os.utime(file_path, (current_time, current_time))
                except (Exception,):
                    logger.exception("Error fixing timestamp for file {}".format(file_path))

    @staticmethod
    def _get_min_timestamp():
        # type: (...) -> float

        min_valid_date = datetime(
            year=1980,
            month=1,
            day=1,
        )
        return time.mktime(min_valid_date.timetuple())

    @staticmethod
    def delete_directory(path):
        # type: (str) -> None

        CommonService.delete_directory(path)

    def extract_zip(self, zip_name, path_to_extract, delete_if_exists=True):
        self.create_dir(path_to_extract, delete_if_exists)

        with zipfile.ZipFile(zip_name, "r") as zip_object:
            zip_object.extractall(path_to_extract)

    @staticmethod
    def extract_tar_gz(file_name, path_to_extract):
        with tarfile.open(file_name, "r:gz") as zip_object:
            zip_object.extractall(path_to_extract)

    @staticmethod
    def rename_folder(original_path, renamed_path):
        if os.path.exists(renamed_path):
            shutil.rmtree(renamed_path, ignore_errors=False, onerror=CommonService.handle_remove_read_only)

        os.rename(original_path, renamed_path)

    def _set_best_id_database_campaign(self, catalog_json):
        # type: (Dict) -> int

        logger.info("Applying: get databases campaign...")
        today = datetime.now()
        databases_ids = dict()
        if catalog_json is None:
            raise ErrorRetrievingJsonCatalog()

        for menu in catalog_json.get("menus"):
            end_in = self._get_date(menu, "endIn", '2099-12-31')
            start_in = self._get_date(menu, "startIn", '1970-01-01')
            if end_in.date() < today.date() or start_in.date() > today.date():
                continue

            databases_ids[menu.get("id")] = abs(start_in - today)

        database_id = min(databases_ids, key=databases_ids.get)
        version = catalog_json["version"]
        update_name = "{}_{}".format(version, database_id)
        last_catalog_id_applied = self._get_last_downloaded_version(UpdateType.catalog)
        last_catalog_id_applied = last_catalog_id_applied.get("updateName")

        if last_catalog_id_applied != update_name:
            self.pos_repository.insert_new_update(database_id, update_name, UpdateType.catalog)
            self.pos_repository.update_step(database_id, UpdateStep.download, UpdateType.catalog)
            logger.info("Applying: get databases campaign finish, catalog to apply: {}...".format(database_id))

        return database_id

    @staticmethod
    def _get_date(menu, key, default):
        # type: (str, str, str) -> datetime

        date = menu.get(key) if menu.get(key, None) is not None else default
        return datetime.strptime(date, "%Y-%m-%d")

    def _get_json_file_databases_campaign(self):
        try:
            with open(self.catalog_json, "r") as catalog_json:
                return json.loads(catalog_json.read().decode('utf-8'))
        except (Exception,):
            logger.info("Error open json file in directory")
            return None

    def update_databases_with_json_config(
        self,               # type: UpdaterService
        update_name,        # type: str
        settings,           # type: settings
    ):
        # type (...) -> None

        settings.set_update_control(
            key=update_name,
            value="$UPDATING_PAYMENTS",
        )
        self._update_payments()

        settings.set_update_control(
            key=update_name,
            value="$UPDATING_NAVIGATIONS",
        )
        self._update_navigations()

    def _update_payments(
        self,       # type: UpdaterService
    ):
        # type (...) -> None

        try:
            logger.info("Updating Payments")

            if not os.path.exists(self.payments_json):
                logger.info("The file {} was not found to update payments database".format(self.payments_json))
                return

            with open(self.payments_json, "r") as payments_json:
                payments = json.loads(payments_json.read())

            all_payments = []
            for payment in payments:
                payment_id = payment.get("id", 0)
                name = payment.get("name", "").encode('utf-8').strip()
                currency = payment.get("tenderCurrency", "BRL")
                change_limit = payment.get("changeLimit", 0)
                electronic_type = payment.get("electronicType", 0)
                open_drawer = payment.get("openDrawer", False)
                need_auth = int(payment.get("needAuth", 0))
                parent_id = payment.get("parentId", None)
                custom_attr = payment.get("customAttr", {})

                link_with_tef = payment.get("linkWithTef", False)
                if link_with_tef:
                    custom_attr["linkWithTef"] = link_with_tef

                payment = Payment(
                    payment_id=payment_id,
                    name=name,
                    currency=currency,
                    change_limit=change_limit,
                    electronic_type=electronic_type,
                    open_drawer=open_drawer,
                    need_auth=need_auth,
                    parent_id=parent_id,
                    custom_attr=json.dumps(custom_attr)
                )
                all_payments.append(payment)

            product_db = os.path.join(self.genesis_databases_path, "product.db")

            if not os.path.exists(product_db):
                raise ErrorGenesisDatabaseNotFound()

            self.common_repository.delete_all_payments(product_db)
            self.common_repository.insert_payments(product_db, all_payments)

            logger.info("Payments updated")
        except ErrorGenesisDatabaseNotFound:
            logger.exception("Error inserting payments on database")
            raise
        except (Exception,):
            logger.exception("Error inserting payments on database")
            raise ErrorUpdatingDatabases()

    def _update_navigations(
        self,       # type: UpdaterService
    ):
        # type (...) -> None

        try:
            logger.info("Updating Navigations")

            product_db = os.path.join(self.genesis_databases_path, "product.db")
            has_necessary_info_to_update_navigations = self._has_necessary_info_to_update_navigations(
                product_db=product_db,
            )
            if not has_necessary_info_to_update_navigations:
                return

            with open(self.navigation_json, "rb") as navigation_json:
                json_navigations = json.loads(navigation_json.read().decode("utf-8"))

            all_navigations = []
            for json_navigation in json_navigations:
                pos_id = json_navigation.get("POSId")
                sale_type = json_navigation.get("SaleType")
                navigation_id = json_navigation.get("NavId")
                if pos_id is None or sale_type is None or navigation_id is None:
                    continue

                json_navigation = PosSaleTypeNavigation(
                    pos_id=pos_id,
                    sale_type=sale_type,
                    navigation_id=navigation_id,
                )
                all_navigations.append(json_navigation)

            self.common_repository.delete_all_pos_navigation_sale_type(
                database_path=product_db,
            )
            self.common_repository.insert_pos_navigation_sale_type(
                database_path=product_db,
                navigations=all_navigations,
            )

            logger.info("Navigations updated")
        except ErrorGenesisDatabaseNotFound:
            logger.exception("Error inserting navigations on database")
            raise
        except (Exception,):
            logger.exception("Error inserting navigations on database")
            raise ErrorUpdatingDatabases()

    def _has_necessary_info_to_update_navigations(
        self,           # type: UpdaterService
        product_db,     # type: str
    ):
        # type: (...) -> bool

        if not os.path.exists(product_db):
            raise ErrorGenesisDatabaseNotFound()

        table_exists = self.common_repository.table_exists_in_db(
            database_path=product_db,
            table_name="PosSaleTypeNavigation",
        )
        if not table_exists:
            return False

        if not os.path.exists(self.navigation_json):
            logger.info("The file {} was not found to update navigation on database".format(self.navigation_json))
            return False

        return True

    @staticmethod
    def get_database_id(database_id):
        # type: (str) -> Optional[str]

        if database_id is None:
            return database_id

        database_id = database_id.split("_")
        position = 1 if len(database_id) > 0 else 0
        return database_id[position]



import json

import cfgtools
from typing import List

from ._PickupArea import PickupArea
from ._PickupAreaSector import PickupAreaSector


def get_first_of(it, default_value=None):
    return next(it, default_value)


class PickupConfig(object):
    def __init__(
        self,           # type: PickupConfig
        pickup_config,  # type: cfgtools.Group
    ):
        # type: (...) -> None

        self.pickup_areas = []
        self.parse_pickup_config(pickup_config=pickup_config)

    def parse_pickup_config(
        self,
        pickup_config,
    ):
        # type: (...) -> None

        for area in pickup_config.groups:
            name = get_first_of(x for x in area.keys if x.name == "PickupName").values[0]
            sectors = get_first_of(x for x in area.keys if x.name == "Sectors").values

            self.pickup_areas.append(
                PickupArea(
                    pickup_name=name,
                    sectors=self._get_sectors(sectors),
                )
            )

    @staticmethod
    def _get_sectors(
        sectors,        # type: List[str]
    ):
        # type: (...) -> List[PickupAreaSector]

        current_sectors = []
        for sector in sectors:
            sector_dict = json.loads(sector)
            pickup_area_sector = PickupAreaSector(
                name=sector_dict.get("name", ""),
                excluded_pos=sector_dict.get("excludedPOS", []),
            )
            current_sectors.append(pickup_area_sector)

        return current_sectors



class NoUpdateFound(Exception):
    pass


class ErrorApplyingUpdate(Exception):
    pass


class ConfigurationReaderException(Exception):
    pass


class ErrorInsertingNewUpdateVersion(Exception):
    pass


class ErrorRetrievingPendingUpdate(Exception):
    pass


class ErrorRetrievingUsers(Exception):
    pass


class ErrorCheckingGenesisDir(Exception):
    pass


class ErrorDownloadingUpdatePackage(Exception):
    pass


class ErrorNotifyingAppliedUpdate(Exception):
    pass


class HasNoProductsImagesToDownload(Exception):
    pass


class NotInsideUpdateWindow(Exception):
    pass


class ErrorCheckingPackageIntegrity(Exception):
    pass


class ErrorUpdatingStep(Exception):
    pass


class ErrorUpdatingStatus(Exception):
    pass


class ErrorRestoringBackup(Exception):
    pass


class ErrorUpdatingStartStop(Exception):
    pass


class ErrorCreatingBackup(Exception):
    pass


class ErrorRestartingHV(Exception):
    pass


class DisabledFunctionality(Exception):
    pass


class ErrorDeletingContent(Exception):
    pass


class ErrorExtractingLoaders(Exception):
    pass


class ErrorRecoveringFolder(Exception):
    pass


class ErrorUpdatingDatabases(Exception):
    pass


class ErrorGenesisDatabaseNotFound(Exception):
    pass


class ErrorDownloadingLoaders(Exception):
    pass


class ErrorDownloadingBinaries(Exception):
    pass


class RollbackError(Exception):
    pass


class InvalidDownloadUpdate(Exception):
    pass


class InvalidBOHConfigurations(Exception):
    pass


class SomeUserLogged(Exception):
    pass


class NotImplementedException(Exception):
    pass


class ErrorGettingTotemConfig(Exception):
    pass


class ErrorRetrievingJsonCatalog(Exception):
    pass


class DataBasesNotFound(Exception):
    pass


class ErrorValidatingOpenedUsers(Exception):
    pass


class ErrorDayIsOpen(Exception):
    pass


class ErrorGettingUsersList(Exception):
    pass


class ErrorUpdatingUserControlType(Exception):
    pass


class NoCatalogDirectoryFoundException(Exception):
    pass


class ErrorUpdatingUserRoles(Exception):
    pass


class NoCatalogJsonFoundException(Exception):
    pass


class NoCatalogUpdateFoundException(Exception):
    pass


class NotAllImagesDownload(Exception):
    pass


class ErrorUpdatingApache(Exception):
    pass


class ErrorIntegrityDatabase(Exception):
    pass


class CannotDeletePdvWithOrders(Exception):
    pass



# ext/preprocessors.py
# Copyright 2006-2024 the Mako authors and contributors <see AUTHORS file>
#
# This module is part of Mako and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

"""preprocessing functions, used with the 'preprocessor'
argument on Template, TemplateLookup"""

import re


def convert_comments(text):
    """preprocess old style comments.

    example:

    from mako.ext.preprocessors import convert_comments
    t = Template(..., preprocessor=convert_comments)"""
    return re.sub(r"(?<=\n)\s*#[^#]", "##", text)



import functools
import weakref

from .__wrapt__ import ObjectProxy, _FunctionWrapperBase

# A weak function proxy. This will work on instance methods, class
# methods, static methods and regular functions. Special treatment is
# needed for the method types because the bound method is effectively a
# transient object and applying a weak reference to one will immediately
# result in it being destroyed and the weakref callback called. The weak
# reference is therefore applied to the instance the method is bound to
# and the original function. The function is then rebound at the point
# of a call via the weak function proxy.

def _weak_function_proxy_callback(ref, proxy, callback):
    if proxy._self_expired:
        return

    proxy._self_expired = True

    # This could raise an exception. We let it propagate back and let
    # the weakref.proxy() deal with it, at which point it generally
    # prints out a short error message direct to stderr and keeps going.

    if callback is not None:
        callback(proxy)

class WeakFunctionProxy(ObjectProxy):

    __slots__ = ('_self_expired', '_self_instance')

    def __init__(self, wrapped, callback=None):
        # We need to determine if the wrapped function is actually a
        # bound method. In the case of a bound method, we need to keep a
        # reference to the original unbound function and the instance.
        # This is necessary because if we hold a reference to the bound
        # function, it will be the only reference and given it is a
        # temporary object, it will almost immediately expire and
        # the weakref callback triggered. So what is done is that we
        # hold a reference to the instance and unbound function and
        # when called bind the function to the instance once again and
        # then call it. Note that we avoid using a nested function for
        # the callback here so as not to cause any odd reference cycles.

        _callback = callback and functools.partial(
                _weak_function_proxy_callback, proxy=self,
                callback=callback)

        self._self_expired = False

        if isinstance(wrapped, _FunctionWrapperBase):
            self._self_instance = weakref.ref(wrapped._self_instance,
                    _callback)

            if wrapped._self_parent is not None:
                super(WeakFunctionProxy, self).__init__(
                        weakref.proxy(wrapped._self_parent, _callback))

            else:
                super(WeakFunctionProxy, self).__init__(
                        weakref.proxy(wrapped, _callback))

            return

        try:
            self._self_instance = weakref.ref(wrapped.__self__, _callback)

            super(WeakFunctionProxy, self).__init__(
                    weakref.proxy(wrapped.__func__, _callback))

        except AttributeError:
            self._self_instance = None

            super(WeakFunctionProxy, self).__init__(
                    weakref.proxy(wrapped, _callback))

    def __call__(*args, **kwargs):
        def _unpack_self(self, *args):
            return self, args

        self, args = _unpack_self(*args)

        # We perform a boolean check here on the instance and wrapped
        # function as that will trigger the reference error prior to
        # calling if the reference had expired.

        instance = self._self_instance and self._self_instance()
        function = self.__wrapped__ and self.__wrapped__

        # If the wrapped function was originally a bound function, for
        # which we retained a reference to the instance and the unbound
        # function we need to rebind the function and then call it. If
        # not just called the wrapped function.

        if instance is None:
            return self.__wrapped__(*args, **kwargs)

        return function.__get__(instance, type(instance))(*args, **kwargs)



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import threading
from enum import Enum

from opentelemetry.instrumentation.utils import http_status_to_status_code
from opentelemetry.semconv.attributes.client_attributes import (
    CLIENT_ADDRESS,
    CLIENT_PORT,
)
from opentelemetry.semconv.attributes.error_attributes import ERROR_TYPE
from opentelemetry.semconv.attributes.http_attributes import (
    HTTP_REQUEST_METHOD,
    HTTP_REQUEST_METHOD_ORIGINAL,
    HTTP_RESPONSE_STATUS_CODE,
    HTTP_ROUTE,
)
from opentelemetry.semconv.attributes.network_attributes import (
    NETWORK_PROTOCOL_VERSION,
)
from opentelemetry.semconv.attributes.server_attributes import (
    SERVER_ADDRESS,
    SERVER_PORT,
)
from opentelemetry.semconv.attributes.url_attributes import (
    URL_FULL,
    URL_PATH,
    URL_QUERY,
    URL_SCHEME,
)
from opentelemetry.semconv.attributes.user_agent_attributes import (
    USER_AGENT_ORIGINAL,
)
from opentelemetry.semconv.trace import SpanAttributes
from opentelemetry.trace.status import Status, StatusCode

# These lists represent attributes for metrics that are currently supported

_client_duration_attrs_old = [
    SpanAttributes.HTTP_STATUS_CODE,
    SpanAttributes.HTTP_HOST,
    SpanAttributes.NET_PEER_PORT,
    SpanAttributes.NET_PEER_NAME,
    SpanAttributes.HTTP_METHOD,
    SpanAttributes.HTTP_FLAVOR,
    SpanAttributes.HTTP_SCHEME,
]

_client_duration_attrs_new = [
    ERROR_TYPE,
    HTTP_REQUEST_METHOD,
    HTTP_RESPONSE_STATUS_CODE,
    NETWORK_PROTOCOL_VERSION,
    SERVER_ADDRESS,
    SERVER_PORT,
    # TODO: Support opt-in for scheme in new semconv
    # URL_SCHEME,
]

_server_duration_attrs_old = [
    SpanAttributes.HTTP_METHOD,
    SpanAttributes.HTTP_HOST,
    SpanAttributes.HTTP_SCHEME,
    SpanAttributes.HTTP_STATUS_CODE,
    SpanAttributes.HTTP_FLAVOR,
    SpanAttributes.HTTP_SERVER_NAME,
    SpanAttributes.NET_HOST_NAME,
    SpanAttributes.NET_HOST_PORT,
]

_server_duration_attrs_new = [
    ERROR_TYPE,
    HTTP_REQUEST_METHOD,
    HTTP_RESPONSE_STATUS_CODE,
    HTTP_ROUTE,
    NETWORK_PROTOCOL_VERSION,
    URL_SCHEME,
]

_server_active_requests_count_attrs_old = [
    SpanAttributes.HTTP_METHOD,
    SpanAttributes.HTTP_HOST,
    SpanAttributes.HTTP_SCHEME,
    SpanAttributes.HTTP_FLAVOR,
    SpanAttributes.HTTP_SERVER_NAME,
]

_server_active_requests_count_attrs_new = [
    HTTP_REQUEST_METHOD,
    URL_SCHEME,
    # TODO: Support SERVER_ADDRESS AND SERVER_PORT
]

OTEL_SEMCONV_STABILITY_OPT_IN = "OTEL_SEMCONV_STABILITY_OPT_IN"


class _OpenTelemetryStabilitySignalType:
    HTTP = "http"


class _HTTPStabilityMode(Enum):
    # http - emit the new, stable HTTP and networking conventions ONLY
    HTTP = "http"
    # http/dup - emit both the old and the stable HTTP and networking conventions
    HTTP_DUP = "http/dup"
    # default - continue emitting old experimental HTTP and networking conventions
    DEFAULT = "default"


def _report_new(mode):
    return mode.name != _HTTPStabilityMode.DEFAULT.name


def _report_old(mode):
    return mode.name != _HTTPStabilityMode.HTTP.name


class _OpenTelemetrySemanticConventionStability:
    _initialized = False
    _lock = threading.Lock()
    _OTEL_SEMCONV_STABILITY_SIGNAL_MAPPING = {}

    @classmethod
    def _initialize(cls):
        with _OpenTelemetrySemanticConventionStability._lock:
            if not _OpenTelemetrySemanticConventionStability._initialized:
                # Users can pass in comma delimited string for opt-in options
                # Only values for http stability are supported for now
                opt_in = os.environ.get(OTEL_SEMCONV_STABILITY_OPT_IN, "")
                opt_in_list = []
                if opt_in:
                    opt_in_list = [s.strip() for s in opt_in.split(",")]
                http_opt_in = _HTTPStabilityMode.DEFAULT
                if opt_in_list:
                    # Process http opt-in
                    # http/dup takes priority over http
                    if _HTTPStabilityMode.HTTP_DUP.value in opt_in_list:
                        http_opt_in = _HTTPStabilityMode.HTTP_DUP
                    elif _HTTPStabilityMode.HTTP.value in opt_in_list:
                        http_opt_in = _HTTPStabilityMode.HTTP
                _OpenTelemetrySemanticConventionStability._OTEL_SEMCONV_STABILITY_SIGNAL_MAPPING[
                    _OpenTelemetryStabilitySignalType.HTTP
                ] = http_opt_in
                _OpenTelemetrySemanticConventionStability._initialized = True

    @classmethod
    # Get OpenTelemetry opt-in mode based off of signal type (http, messaging, etc.)
    def _get_opentelemetry_stability_opt_in_mode(
        cls,
        signal_type: _OpenTelemetryStabilitySignalType,
    ) -> _HTTPStabilityMode:
        return _OpenTelemetrySemanticConventionStability._OTEL_SEMCONV_STABILITY_SIGNAL_MAPPING.get(
            signal_type, _HTTPStabilityMode.DEFAULT
        )


def _filter_semconv_duration_attrs(
    attrs,
    old_attrs,
    new_attrs,
    sem_conv_opt_in_mode=_HTTPStabilityMode.DEFAULT,
):
    filtered_attrs = {}
    # duration is two different metrics depending on sem_conv_opt_in_mode, so no DUP attributes
    allowed_attributes = (
        new_attrs
        if sem_conv_opt_in_mode == _HTTPStabilityMode.HTTP
        else old_attrs
    )
    for key, val in attrs.items():
        if key in allowed_attributes:
            filtered_attrs[key] = val
    return filtered_attrs


def _filter_semconv_active_request_count_attr(
    attrs,
    old_attrs,
    new_attrs,
    sem_conv_opt_in_mode=_HTTPStabilityMode.DEFAULT,
):
    filtered_attrs = {}
    if _report_old(sem_conv_opt_in_mode):
        for key, val in attrs.items():
            if key in old_attrs:
                filtered_attrs[key] = val
    if _report_new(sem_conv_opt_in_mode):
        for key, val in attrs.items():
            if key in new_attrs:
                filtered_attrs[key] = val
    return filtered_attrs


def set_string_attribute(result, key, value):
    if value:
        result[key] = value


def set_int_attribute(result, key, value):
    if value:
        try:
            result[key] = int(value)
        except ValueError:
            return


def _set_http_method(result, original, normalized, sem_conv_opt_in_mode):
    original = original.strip()
    normalized = normalized.strip()
    # See https://github.com/open-telemetry/semantic-conventions/blob/main/docs/http/http-spans.md#common-attributes
    # Method is case sensitive. "http.request.method_original" should not be sanitized or automatically capitalized.
    if original != normalized and _report_new(sem_conv_opt_in_mode):
        set_string_attribute(result, HTTP_REQUEST_METHOD_ORIGINAL, original)

    if _report_old(sem_conv_opt_in_mode):
        set_string_attribute(result, SpanAttributes.HTTP_METHOD, normalized)
    if _report_new(sem_conv_opt_in_mode):
        set_string_attribute(result, HTTP_REQUEST_METHOD, normalized)


def _set_http_status_code(result, code, sem_conv_opt_in_mode):
    if _report_old(sem_conv_opt_in_mode):
        set_int_attribute(result, SpanAttributes.HTTP_STATUS_CODE, code)
    if _report_new(sem_conv_opt_in_mode):
        set_int_attribute(result, HTTP_RESPONSE_STATUS_CODE, code)


def _set_http_url(result, url, sem_conv_opt_in_mode):
    if _report_old(sem_conv_opt_in_mode):
        set_string_attribute(result, SpanAttributes.HTTP_URL, url)
    if _report_new(sem_conv_opt_in_mode):
        set_string_attribute(result, URL_FULL, url)


def _set_http_scheme(result, scheme, sem_conv_opt_in_mode):
    if _report_old(sem_conv_opt_in_mode):
        set_string_attribute(result, SpanAttributes.HTTP_SCHEME, scheme)
    if _report_new(sem_conv_opt_in_mode):
        set_string_attribute(result, URL_SCHEME, scheme)


def _set_http_flavor_version(result, version, sem_conv_opt_in_mode):
    if _report_old(sem_conv_opt_in_mode):
        set_string_attribute(result, SpanAttributes.HTTP_FLAVOR, version)
    if _report_new(sem_conv_opt_in_mode):
        set_string_attribute(result, NETWORK_PROTOCOL_VERSION, version)


def _set_http_user_agent(result, user_agent, sem_conv_opt_in_mode):
    if _report_old(sem_conv_opt_in_mode):
        set_string_attribute(
            result, SpanAttributes.HTTP_USER_AGENT, user_agent
        )
    if _report_new(sem_conv_opt_in_mode):
        set_string_attribute(result, USER_AGENT_ORIGINAL, user_agent)


# Client


def _set_http_host_client(result, host, sem_conv_opt_in_mode):
    if _report_old(sem_conv_opt_in_mode):
        set_string_attribute(result, SpanAttributes.HTTP_HOST, host)
    if _report_new(sem_conv_opt_in_mode):
        set_string_attribute(result, SERVER_ADDRESS, host)


def _set_http_net_peer_name_client(result, peer_name, sem_conv_opt_in_mode):
    if _report_old(sem_conv_opt_in_mode):
        set_string_attribute(result, SpanAttributes.NET_PEER_NAME, peer_name)
    if _report_new(sem_conv_opt_in_mode):
        set_string_attribute(result, SERVER_ADDRESS, peer_name)


def _set_http_peer_port_client(result, port, sem_conv_opt_in_mode):
    if _report_old(sem_conv_opt_in_mode):
        set_int_attribute(result, SpanAttributes.NET_PEER_PORT, port)
    if _report_new(sem_conv_opt_in_mode):
        set_int_attribute(result, SERVER_PORT, port)


def _set_http_network_protocol_version(result, version, sem_conv_opt_in_mode):
    if _report_old(sem_conv_opt_in_mode):
        set_string_attribute(result, SpanAttributes.HTTP_FLAVOR, version)
    if _report_new(sem_conv_opt_in_mode):
        set_string_attribute(result, NETWORK_PROTOCOL_VERSION, version)


# Server


def _set_http_net_host(result, host, sem_conv_opt_in_mode):
    if _report_old(sem_conv_opt_in_mode):
        set_string_attribute(result, SpanAttributes.NET_HOST_NAME, host)
    if _report_new(sem_conv_opt_in_mode):
        set_string_attribute(result, SERVER_ADDRESS, host)


def _set_http_net_host_port(result, port, sem_conv_opt_in_mode):
    if _report_old(sem_conv_opt_in_mode):
        set_int_attribute(result, SpanAttributes.NET_HOST_PORT, port)
    if _report_new(sem_conv_opt_in_mode):
        set_int_attribute(result, SERVER_PORT, port)


def _set_http_target(result, target, path, query, sem_conv_opt_in_mode):
    if _report_old(sem_conv_opt_in_mode):
        set_string_attribute(result, SpanAttributes.HTTP_TARGET, target)
    if _report_new(sem_conv_opt_in_mode):
        if path:
            set_string_attribute(result, URL_PATH, path)
        if query:
            set_string_attribute(result, URL_QUERY, query)


def _set_http_host_server(result, host, sem_conv_opt_in_mode):
    if _report_old(sem_conv_opt_in_mode):
        set_string_attribute(result, SpanAttributes.HTTP_HOST, host)
    if _report_new(sem_conv_opt_in_mode):
        set_string_attribute(result, CLIENT_ADDRESS, host)


# net.peer.ip -> net.sock.peer.addr
# https://github.com/open-telemetry/semantic-conventions/blob/40db676ca0e735aa84f242b5a0fb14e49438b69b/schemas/1.15.0#L18
# net.sock.peer.addr -> client.socket.address for server spans (TODO) AND client.address if missing
# https://github.com/open-telemetry/semantic-conventions/blob/v1.21.0/CHANGELOG.md#v1210-2023-07-13
# https://github.com/open-telemetry/semantic-conventions/blob/main/docs/non-normative/http-migration.md#common-attributes-across-http-client-and-server-spans
def _set_http_peer_ip_server(result, ip, sem_conv_opt_in_mode):
    if _report_old(sem_conv_opt_in_mode):
        set_string_attribute(result, SpanAttributes.NET_PEER_IP, ip)
    if _report_new(sem_conv_opt_in_mode):
        # Only populate if not already populated
        if not result.get(CLIENT_ADDRESS):
            set_string_attribute(result, CLIENT_ADDRESS, ip)


def _set_http_peer_port_server(result, port, sem_conv_opt_in_mode):
    if _report_old(sem_conv_opt_in_mode):
        set_int_attribute(result, SpanAttributes.NET_PEER_PORT, port)
    if _report_new(sem_conv_opt_in_mode):
        set_int_attribute(result, CLIENT_PORT, port)


def _set_http_net_peer_name_server(result, name, sem_conv_opt_in_mode):
    if _report_old(sem_conv_opt_in_mode):
        set_string_attribute(result, SpanAttributes.NET_PEER_NAME, name)
    if _report_new(sem_conv_opt_in_mode):
        set_string_attribute(result, CLIENT_ADDRESS, name)


def _set_status(
    span,
    metrics_attributes: dict,
    status_code: int,
    status_code_str: str,
    server_span: bool = True,
    sem_conv_opt_in_mode: _HTTPStabilityMode = _HTTPStabilityMode.DEFAULT,
):
    if status_code < 0:
        metrics_attributes[ERROR_TYPE] = status_code_str
        if span.is_recording():
            if _report_new(sem_conv_opt_in_mode):
                span.set_attribute(ERROR_TYPE, status_code_str)
            span.set_status(
                Status(
                    StatusCode.ERROR,
                    "Non-integer HTTP status: " + status_code_str,
                )
            )
    else:
        status = http_status_to_status_code(
            status_code, server_span=server_span
        )

        if _report_old(sem_conv_opt_in_mode):
            if span.is_recording():
                span.set_attribute(
                    SpanAttributes.HTTP_STATUS_CODE, status_code
                )
            metrics_attributes[SpanAttributes.HTTP_STATUS_CODE] = status_code
        if _report_new(sem_conv_opt_in_mode):
            if span.is_recording():
                span.set_attribute(HTTP_RESPONSE_STATUS_CODE, status_code)
            metrics_attributes[HTTP_RESPONSE_STATUS_CODE] = status_code
            if status == StatusCode.ERROR:
                if span.is_recording():
                    span.set_attribute(ERROR_TYPE, status_code_str)
                metrics_attributes[ERROR_TYPE] = status_code_str
        if span.is_recording():
            span.set_status(Status(status))


# Get schema version based off of opt-in mode
def _get_schema_url(mode: _HTTPStabilityMode) -> str:
    if mode is _HTTPStabilityMode.DEFAULT:
        return "https://opentelemetry.io/schemas/1.11.0"
    return SpanAttributes.SCHEMA_URL



from ._CatalogUpdaterProcessor import CatalogUpdaterProcessor
from ._MediaUpdaterProcessor import MediaUpdaterProcessor
from ._UserUpdaterProcessor import UserUpdaterProcessor
from ._GetCatalogVersionToApplyProcessor import GetCatalogVersionToApplyProcessor
from ._LoaderUpdaterProcessor import LoaderUpdaterProcessor
from ._RemoteLoaderUpdaterProcessor import RemoteLoaderUpdaterProcessor
from ._GetVersionsProcessor import GetVersionsProcessor
from ._BinaryUpdaterProcessor import BinaryUpdaterProcessor
from ._GetBinaryUpdaterProcessor import GetBinaryUpdaterProcessor
from ._ChromeUpdaterProcessor import ChromeUpdaterProcessor
from ._BaseMessageProcessor import BaseUpdaterProcessor
from ._DefaultMessageProcessor import DefaultMessageProcessor
from ._HasPendingLoaderUpdateProcessor import HasPendingLoaderUpdateProcessor



from ._AddTagBox import AddTagBox
from ._AddTimeTagBox import AddTimeTagBox
from ._AllDoneBlinkBox import AllDoneBlinkBox
from ._AllItemHasAtLeastOneTagFilterBox import AllItemHasAtLeastOneTagFilterBox
from ._AllItemsDoneBox import AllItemsDoneBox
from ._AllServedItemsBox import AllServedItemsBox
from ._AllowedForbiddenFilter import AllowedForbiddenFilter
from ._CFHHiderBox import CFHHiderBox
from ._CollapseSameItemsBox import CollapseSameItemsBox
from ._ComboDoneWhenAllDoneBox import ComboDoneWhenAllDoneBox
from ._ComboFilterBox import ComboFilterBox
from ._ComboServedWhenAllServedBox import ComboServedWhenAllServedBox
from ._ComboTaggedWhenAllHasTagBox import ComboTaggedWhenAllHasTagBox
from ._CourseBox import CourseBox
from ._CustomParamCourseBox import CustomParamCourseBox
from ._CustomParamPrepTimeBox import CustomParamPrepTimeBox
from ._CustomPropertyFilterBox import CustomPropertyFilterBox
from ._CustomPropertyValueFilterBox import CustomPropertyValueFilterBox
from ._DecreaseTimeTagBox import DecreaseTimeTagBox
from ._DisplayTimeFilterBox import DisplayTimeFilterBox
from ._DoneFilterBox import DoneFilterBox
from ._DoneWhenAnyDoneBox import DoneWhenAnyDoneBox
from ._FilterDefaultQuantityIngredientsBox import FilterDefaultQuantityIngredientsBox
from ._FilterDeletedItemsBeforeFirstStateBox import FilterDeletedItemsBeforeFirstStateBox
from ._FilterDeletedItemsNotSentToKitchenBox import FilterDeletedItemsNotSentToKitchenBox
from ._FilterOrderIfAllItemsHaveTagBox import FilterOrderIfAllItemsHaveTagBox
from ._FilterOrderIfAtLeastOneItemHasAtLeastOneTagFilterBox import FilterOrderIfAtLeastOneItemHasAtLeastOneTagFilterBox
from ._FilterOrderIfNotAllItemsHaveTagBox import FilterOrderIfNotAllItemsHaveTagBox
from ._FromOrderStateFilterBox import FromOrderStateFilterBox
from ._HideOptionBox import HideOptionBox
from ._ItemCustomParamsFilterBox import ItemCustomParamsFilterBox
from ._ItemFilterBox import ItemFilterBox
from ._ItemPropertiesFilterBox import ItemPropertiesFilterBox
from ._ItemSequencerBox import ItemSequencerBox
from ._ItemSorterBox import ItemSorterBox
from ._JitAddTagBox import JitAddTagBox
from ._JitDoneWhenAnyDoneBox import JitDoneWhenAnyDoneBox
from ._JitItemFilterBox import JitItemFilterBox
from ._JitSequencerItemFilterBox import JitSequencerItemFilterBox
from ._KeepProdStateBox import KeepProdStateBox
from ._LastTagTimerBox import LastTagTimerBox
from ._NoJitAddTagBox import NoJitAddTagBox
from ._NonComboFilterBox import NonComboFilterBox
from ._NonValidJitAddTagBox import NonValidJitAddTagBox
from ._NotShowOnKitchenAddTagBox import NotShowOnKitchenAddTagBox
from ._OrderProductCategoryFilterBox import OrderProductCategoryFilterBox
from ._OrderPropertiesFilterBox import OrderPropertiesFilterBox
from ._OrderRoundRobinBox import OrderRoundRobinBox
from ._OrderSequencerBox import OrderSequencerBox
from ._OrderSpreaderBox import OrderSpreaderBox
from ._OrderStateFilterBox import OrderStateFilterBox
from ._OrderStateTimerBox import OrderStateTimerBox
from ._OrderSubTypeFilterBox import OrderSubTypeFilterBox
from ._PodTypeFilterBox import PodTypeFilterBox
from ._OrderCostAdderBox import OrderCostAdderBox
from ._PrepTimeBox import PrepTimeBox
from ._ProdStateFilterBox import ProdStateFilterBox
from ._ProdStateTimerBox import ProdStateTimerBox
from ._ProductCodeFilterBox import ProductCodeFilterBox
from ._ProductionBox import ProductionBox
from ._RecalledOrderStampBox import RecalledOrderStampBox
from ._RemoveDeletedItemsBox import RemoveDeletedItemsBox
from ._RemoveDoingWhenDoneBox import RemoveDoingWhenDoneBox
from ._RemoveReprocessedOrdersBox import RemoveReprocessedOrdersBox
from ._ReprocessOrdersBox import ReprocessOrdersBox
from ._SaleTypeFilterBox import SaleTypeFilterBox
from ._SendMomentTimerBox import SendMomentTimerBox
from ._SequencerTagItemBox import SequencerTagItemBox
from ._ShowOnKitchenFilterBox import ShowOnKitchenFilterBox
from ._SkipCourseBox import SkipCourseBox
from ._SorterBox import SorterBox
from ._StampItemsViewBox import StampItemsViewBox
from ._TagFilterBox import TagFilterBox
from ._TagItemsWhenProdStateBox import TagItemsWhenProdStateBox
from ._ViewBox import ViewBox
from ._AddAllItemsTagBox import AddAllItemsTagBox



# -*- coding: utf-8 -*-

from sysactions import action, show_messagebox  # noqa

from msgbus import TK_SYS_NAK
from api.pos_updater import get_pos_updater_service
from messagebus import Message
from .update_util import show_nak_message
from ..util.get_menu_user_logged import get_menu_manager_authenticated_user


@action
def update_loaders(
    pos_id,    # type: str
):
    # type: (...) -> None

    resp = show_messagebox(
        posid=pos_id,
        message="$UPDATE_LOADERS_MESSAGE_QUESTION",
        buttons="$OK|$CANCEL",
    )
    if resp == 1:
        return

    msg = _send_update_message(pos_id=pos_id)
    if msg.token == TK_SYS_NAK:
        response_message = msg.data
        show_nak_message(
            pos_id=pos_id,
            update_key="$UPDATE_LOADERS_",
            message=response_message,
        )


def send_update_loaders_with_retry(
    pos_id,    # type: str
):
    # type: (...) -> None

    while True:
        msg = _send_update_message(pos_id=pos_id)
        if msg.token == TK_SYS_NAK:
            resp = show_messagebox(
                posid=pos_id,
                message="$UPDATE_LOADERS_ERROR_TRY_AGAIN",
                buttons="$YES|$NO",
            )
            if resp == 1:
                return


def _send_update_message(
    pos_id,     # type: str
):
    # type: (...) -> Message

    user_id = get_menu_manager_authenticated_user(pos_id=int(pos_id))
    updater_service = get_pos_updater_service(pos_id=pos_id)
    return updater_service.loader_update(user_id=user_id)



import json
import logging

from .dto import LockBenefitRequest
from application.model import ApplierSubTypes
from application.repository.api import APIRepository
from order_api import OrderService
from order_api.dto import CustomProperty

logger = logging.getLogger(__name__)


class LockBenefitInteractor(object):
    def __init__(
        self,            # type: LockBenefitInteractor
        api_repository,  # type: APIRepository
    ):
        # type: (...) -> None

        self.api_repository = api_repository

    def lock_benefit(
        self,       # type: LockBenefitInteractor
        request,    # type: LockBenefitRequest
    ):
        # type: (...) -> str

        if request.benefit.get("applierSubType") == ApplierSubTypes.LOYALTY:
            self._update_applied_loyalty_on_total(
                pos_id=request.pos_id,
                benefit_id=request.benefit["id"],
            )

        logger.info("Locking benefit: {}; PosId: {}".format(request.benefit.get("id"), request.pos_id))
        lock_id = self.api_repository.lock_benefit(
            benefit=request.benefit,
            pos_id=request.pos_id,
            payment_info=request.payment_info,
        )
        message = "Benefit {}; PosId: {} locked with lock id: {}"
        logger.info(message.format(request.benefit.get("id"), request.pos_id, lock_id))
        return lock_id

    def _update_applied_loyalty_on_total(
        self,           # type: LockBenefitInteractor
        pos_id,         # type: int
        benefit_id,     # type: str
    ):
        # type: (...) -> None

        order_service = OrderService(
            pos_id=pos_id,
            message_bus=self.api_repository.message_bus,
        )
        order_picture = order_service.get_order_picture()
        prop = order_picture.find(".//OrderProperty[@key='LOYALTY_COUPONS_TO_APPLY_ON_TOTAL']")
        loyalty_on_total = json.loads(prop.get("value") or "[]") if prop is not None else []

        if benefit_id not in loyalty_on_total:
            loyalty_on_total.append(benefit_id)

        order_service.set_order_custom_property(custom_property=CustomProperty(
            key="LOYALTY_COUPONS_TO_APPLY_ON_TOTAL",
            value=json.dumps(loyalty_on_total),
        ))



import os
import sqlite3
from sqlite3 import Connection


class SqliteRepositoryFactory(object):

    def __init__(
        self,
        database_dir: str,
    ) -> None:

        self.database_dir = database_dir

    def create_connection(
        self,
        db_name: str,
    ) -> Connection:

        db_path = os.path.abspath(os.path.join(self.database_dir, db_name))
        conn_string = f"file:///{db_path}.db"
        return sqlite3.connect(database=conn_string, uri=True)



class GetTotaledMessageRequest(object):
    def __init__(
        self,
        pos_id,         # type: str
        customer_id,    # type: str
    ):
        # type: (...) -> None

        self.pos_id = int(pos_id)
        self.customer_id = customer_id



from bustoken import create_token, MSGPRT_LOW
from msgbus import TK_SYS_NAK, MBException
from sysactions import action, send_message, show_messagebox, close_asynch_dialog

TK_POS_UPDATER_PERFORM_USER_UPDATE = create_token(MSGPRT_LOW, "95", "3")


@action
def importEmployees(pos_id):
    wait_dlg_id = show_messagebox(pos_id, "$PLEASE_WAIT", "", buttons="", asynch=True)
    success = False
    try:
        reply = send_message("PosUpdater", TK_POS_UPDATER_PERFORM_USER_UPDATE, timeout=30000000)
        if reply.token == TK_SYS_NAK:
            raise MBException(reply.data, "")
        success = True
        return True
    except MBException as _:
        return False
    finally:
        close_asynch_dialog(pos_id, wait_dlg_id)
        if success:
            show_messagebox(pos_id, title="", message="$OPERATION_SUCCEEDED")
        else:
            show_messagebox(pos_id, title="", message="$OPERATION_FAILED")



from typing import Any
from typing import Optional


class Timeout(object):
    def __init__(
        self,                   # type: Timeout
        minutes=None,           # type: Optional[int]
        seconds=None,           # type: Optional[int]
        milliseconds=None,      # type: Optional[int]
        microseconds=None,      # type: Optional[int]
    ):
        # type: (...) -> None

        if minutes is None and seconds is None and milliseconds is None and microseconds is None:
            raise RuntimeError("At least one time field must be supplied")

        self.minutes = minutes if minutes is not None else 0
        self.seconds = seconds if seconds is not None else 0
        self.milliseconds = milliseconds if milliseconds is not None else 0
        self.microseconds = microseconds if microseconds is not None else 0

    def __eq__(
        self,       # type: Timeout
        other,      # type: Any
    ):
        # type: (...) -> bool

        if not isinstance(other, Timeout):
            return False

        return (
            self.minutes == other.minutes and
            self.seconds == other.seconds and
            self.milliseconds == other.milliseconds and
            self.microseconds == other.microseconds
        )

    def __ne__(
        self,       # type: Timeout
        other,      # type: Any
    ):
        # type: (...) -> bool

        return not self.__eq__(other)

    def __str__(
        self,       # type: Timeout
    ):
        # type: (...) -> str

        return str({
            "minutes": self.minutes,
            "seconds": self.seconds,
            "milliseconds": self.milliseconds,
            "microseconds": self.microseconds
        })



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2019, OpenTelemetry Authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import opentelemetry.proto.trace.v1.trace_pb2
import sys

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing_extensions.final
class ExportTraceServiceRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    RESOURCE_SPANS_FIELD_NUMBER: builtins.int
    @property
    def resource_spans(
        self,
    ) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[
        opentelemetry.proto.trace.v1.trace_pb2.ResourceSpans
    ]:
        """An array of ResourceSpans.
        For data coming from a single resource this array will typically contain one
        element. Intermediary nodes (such as OpenTelemetry Collector) that receive
        data from multiple origins typically batch the data before forwarding further and
        in that case this array will contain multiple elements.
        """

    def __init__(
        self,
        *,
        resource_spans: (
            collections.abc.Iterable[
                opentelemetry.proto.trace.v1.trace_pb2.ResourceSpans
            ]
            | None
        ) = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "resource_spans", b"resource_spans"
        ],
    ) -> None: ...

global___ExportTraceServiceRequest = ExportTraceServiceRequest

@typing_extensions.final
class ExportTraceServiceResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    PARTIAL_SUCCESS_FIELD_NUMBER: builtins.int
    @property
    def partial_success(self) -> global___ExportTracePartialSuccess:
        """The details of a partially successful export request.

        If the request is only partially accepted
        (i.e. when the server accepts only parts of the data and rejects the rest)
        the server MUST initialize the `partial_success` field and MUST
        set the `rejected_<signal>` with the number of items it rejected.

        Servers MAY also make use of the `partial_success` field to convey
        warnings/suggestions to senders even when the request was fully accepted.
        In such cases, the `rejected_<signal>` MUST have a value of `0` and
        the `error_message` MUST be non-empty.

        A `partial_success` message with an empty value (rejected_<signal> = 0 and
        `error_message` = "") is equivalent to it not being set/present. Senders
        SHOULD interpret it the same way as in the full success case.
        """

    def __init__(
        self,
        *,
        partial_success: global___ExportTracePartialSuccess | None = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "partial_success", b"partial_success"
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "partial_success", b"partial_success"
        ],
    ) -> None: ...

global___ExportTraceServiceResponse = ExportTraceServiceResponse

@typing_extensions.final
class ExportTracePartialSuccess(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    REJECTED_SPANS_FIELD_NUMBER: builtins.int
    ERROR_MESSAGE_FIELD_NUMBER: builtins.int
    rejected_spans: builtins.int
    """The number of rejected spans.

    A `rejected_<signal>` field holding a `0` value indicates that the
    request was fully accepted.
    """
    error_message: builtins.str
    """A developer-facing human-readable message in English. It should be used
    either to explain why the server rejected parts of the data during a partial
    success or to convey warnings/suggestions during a full success. The message
    should offer guidance on how users can address such issues.

    error_message is an optional field. An error_message with an empty value
    is equivalent to it not being set.
    """
    def __init__(
        self,
        *,
        rejected_spans: builtins.int = ...,
        error_message: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "error_message",
            b"error_message",
            "rejected_spans",
            b"rejected_spans",
        ],
    ) -> None: ...

global___ExportTracePartialSuccess = ExportTracePartialSuccess



from typing import Optional


class UserDto(object):

    def __init__(
        self,
        user_id,                    # type: int
        level,                      # type: int
        user_name,                  # type: str
        long_name,                  # type: str
        status,                     # type: int
        admission_date=None,        # type: Optional[str]
        termination_date=None,      # type: Optional[str]
        password=None,              # type: Optional[str]
        fingerprint_minutiae=None,  # type: Optional[str]
        nfc_uuid=None,              # type: Optional[str]
    ):
        # type: (...) -> None

        self.user_id = user_id
        self.level = int(level)
        self.user_name = user_name
        self.long_name = long_name
        self.status = int(status) if status is not None else status
        self.admission_date = admission_date
        self.termination_date = termination_date
        self.password = password
        self.fingerprint_minutiae = fingerprint_minutiae
        self.nfc_uuid = nfc_uuid

    def authenticate_response(
        self,   # type: UserDto
    ):
        # type: (...) -> bytes

        user_params_data = "{}\0{}\0{}\0{}\0{}\0{}".format(
            self.level,
            self.user_name,
            self.long_name,
            self.status,
            self.admission_date or "",
            self.termination_date or "",
        )
        return user_params_data.encode("utf-8")

    def authenticate_nfc_response(
        self,   # type: UserDto
    ):
        # type: (...) -> bytes

        user_params_data = "{}\0{}\0{}\0{}\0{}\0{}\0{}".format(
            self.user_id,
            self.level,
            self.user_name,
            self.long_name,
            self.status,
            self.admission_date or "",
            self.termination_date or "",
        )
        return user_params_data.encode("utf-8")

    def get_info_response(
        self,   # type: UserDto
    ):
        # type: (...) -> str

        user_attributes = (
            'UserId="{0}" '
            'UserName="{1}" '
            'LongName="{2}" '
            'Level="{3}" '
            'Status="{4}"'
        ).format(
            self.user_id,
            self.user_name,
            self.long_name,
            self.level,
            self.status,
        )

        if self.admission_date:
            user_attributes += ' AdmissionDate="{}"'.format(self.admission_date)
        if self.termination_date:
            user_attributes += ' TerminationDate="{}"'.format(self.termination_date)

        user_event_data = "<user {}/>".format(user_attributes)
        return user_event_data



# unp.py - functions for handling Belarusian UNP numbers
# coding: utf-8
#
# Copyright (C) 2020 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""УНП, UNP (Учетный номер плательщика, the Belarus VAT number).

The УНП (UNP) or Учетный номер плательщика (Uchetniy nomer platel'shika,
Payer account number) is issued to organisations and individuals for tax
purposes. The number consists of 9 digits (numeric for organisations,
alphanumeric for individuals) and contains a region identifier, a serial per
region and a check digit.

More information:

* https://be.wikipedia.org/wiki/Уліковы_нумар_плацельшчыка
* http://pravo.levonevsky.org/bazaby09/sbor37/text37892/index3.htm

>>> validate('200988541')
'200988541'
>>> validate('УНП MA1953684')
'MA1953684'
>>> validate('200988542')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
"""

from stdnum.exceptions import *
from stdnum.util import clean, isdigits, to_unicode


# Mapping of Cyrillic letters to Latin letters
_cyrillic_to_latin = dict(zip(
    u'АВЕКМНОРСТ',
    u'ABEKMHOPCT',
))


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    number = clean(number, ' ').upper().strip()
    for prefix in ('УНП', u'УНП', 'UNP', u'UNP'):
        if type(number) == type(prefix) and number.startswith(prefix):
            number = number[len(prefix):]
    # Replace Cyrillic letters with Latin letters
    cleaned = ''.join(_cyrillic_to_latin.get(x, x) for x in to_unicode(number))
    if type(cleaned) != type(number):  # pragma: no cover (Python2 only)
        cleaned = cleaned.encode('utf-8')
    return cleaned


def calc_check_digit(number):
    """Calculate the check digit for the number."""
    number = compact(number)
    alphabet = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ'
    weights = (29, 23, 19, 17, 13, 7, 5, 3)
    if not isdigits(number):
        number = number[0] + str('ABCEHKMOPT'.index(number[1])) + number[2:]
    c = sum(w * alphabet.index(n) for w, n in zip(weights, number)) % 11
    if c > 9:
        raise InvalidChecksum()
    return str(c)


def validate(number):
    """Check if the number is a valid number. This checks the length,
    formatting and check digit."""
    number = compact(number)
    if len(number) != 9:
        raise InvalidLength()
    if not isdigits(number[2:]):
        raise InvalidFormat()
    if not isdigits(number[:2]) and not all(x in 'ABCEHKMOPT' for x in number[:2]):
        raise InvalidFormat()
    if number[0] not in '1234567ABCEHKM':
        raise InvalidComponent()
    if number[-1] != calc_check_digit(number):
        raise InvalidChecksum()
    return number


def is_valid(number):
    """Check if the number is a valid number."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def check_nalog(number, timeout=30):  # pragma: no cover (not part of normal test suite)
    """Retrieve registration information from the portal.nalog.gov.by web site.

    This basically returns the JSON response from the web service as a dict.
    Will return ``None`` if the number is invalid or unknown.
    """
    # this function isn't automatically tested because it would require
    # network access for the tests and unnecessarily load the web service
    # Since the nalog.gov.by web site currently provides an incomplete
    # certificate chain, we provide our own.
    import requests
    response = requests.get(
        'https://www.portal.nalog.gov.by/grp/getData',
        params={
            'unp': compact(number),
            'charset': 'UTF-8',
            'type': 'json'},
        timeout=timeout)
    if response.ok and response.content:
        return response.json()['row']



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from typing import Final

from deprecated import deprecated

TELEMETRY_DISTRO_NAME: Final = "telemetry.distro.name"
"""
The name of the auto instrumentation agent or distribution, if used.
Note: Official auto instrumentation agents and distributions SHOULD set the `telemetry.distro.name` attribute to
a string starting with `opentelemetry-`, e.g. `opentelemetry-java-instrumentation`.
"""

TELEMETRY_DISTRO_VERSION: Final = "telemetry.distro.version"
"""
The version string of the auto instrumentation agent or distribution, if used.
"""

TELEMETRY_SDK_LANGUAGE: Final = "telemetry.sdk.language"
"""
Deprecated in favor of stable :py:const:`opentelemetry.semconv.attributes.telemetry_attributes.TELEMETRY_SDK_LANGUAGE`.
"""

TELEMETRY_SDK_NAME: Final = "telemetry.sdk.name"
"""
Deprecated in favor of stable :py:const:`opentelemetry.semconv.attributes.telemetry_attributes.TELEMETRY_SDK_NAME`.
"""

TELEMETRY_SDK_VERSION: Final = "telemetry.sdk.version"
"""
Deprecated in favor of stable :py:const:`opentelemetry.semconv.attributes.telemetry_attributes.TELEMETRY_SDK_VERSION`.
"""


@deprecated(
    reason="Deprecated in favor of stable :py:const:`opentelemetry.semconv.attributes.telemetry_attributes.TelemetrySdkLanguageValues`."
)  # type: ignore
class TelemetrySdkLanguageValues(Enum):
    CPP = "cpp"
    """Deprecated in favor of stable :py:const:`opentelemetry.semconv.attributes.telemetry_attributes.TelemetrySdkLanguageValues.CPP`."""
    DOTNET = "dotnet"
    """Deprecated in favor of stable :py:const:`opentelemetry.semconv.attributes.telemetry_attributes.TelemetrySdkLanguageValues.DOTNET`."""
    ERLANG = "erlang"
    """Deprecated in favor of stable :py:const:`opentelemetry.semconv.attributes.telemetry_attributes.TelemetrySdkLanguageValues.ERLANG`."""
    GO = "go"
    """Deprecated in favor of stable :py:const:`opentelemetry.semconv.attributes.telemetry_attributes.TelemetrySdkLanguageValues.GO`."""
    JAVA = "java"
    """Deprecated in favor of stable :py:const:`opentelemetry.semconv.attributes.telemetry_attributes.TelemetrySdkLanguageValues.JAVA`."""
    NODEJS = "nodejs"
    """Deprecated in favor of stable :py:const:`opentelemetry.semconv.attributes.telemetry_attributes.TelemetrySdkLanguageValues.NODEJS`."""
    PHP = "php"
    """Deprecated in favor of stable :py:const:`opentelemetry.semconv.attributes.telemetry_attributes.TelemetrySdkLanguageValues.PHP`."""
    PYTHON = "python"
    """Deprecated in favor of stable :py:const:`opentelemetry.semconv.attributes.telemetry_attributes.TelemetrySdkLanguageValues.PYTHON`."""
    RUBY = "ruby"
    """Deprecated in favor of stable :py:const:`opentelemetry.semconv.attributes.telemetry_attributes.TelemetrySdkLanguageValues.RUBY`."""
    RUST = "rust"
    """Deprecated in favor of stable :py:const:`opentelemetry.semconv.attributes.telemetry_attributes.TelemetrySdkLanguageValues.RUST`."""
    SWIFT = "swift"
    """Deprecated in favor of stable :py:const:`opentelemetry.semconv.attributes.telemetry_attributes.TelemetrySdkLanguageValues.SWIFT`."""
    WEBJS = "webjs"
    """Deprecated in favor of stable :py:const:`opentelemetry.semconv.attributes.telemetry_attributes.TelemetrySdkLanguageValues.WEBJS`."""



from .test_autogen_comments import *  # noqa
from .test_autogen_computed import *  # noqa
from .test_autogen_diffs import *  # noqa
from .test_autogen_fks import *  # noqa
from .test_autogen_identity import *  # noqa
from .test_environment import *  # noqa
from .test_op import *  # noqa



import json
import logging
from datetime import (
    datetime,
    timedelta,
)
from threading import Lock

from typing import Optional
from xml.etree import cElementTree as eTree

from bustoken import (
    TK_LOYALTY_GET_ID_TYPE,
    TK_LOYALTY_GET_LOYALTY_ID,
    TK_LOYALTY_IS_AVAILABLE,
    TK_LOYALTY_ID_NOT_FOUND,
)
from msgbus import (
    FM_PARAM,
    MBException,
    TK_SYS_ACK,
)
from sysactions import (
    send_message,
    show_messagebox,
)

from mw_helper import show_qr_code

from utilfunctions import (
    get_custom_property_value,
    parse_qr_tag,
)

show_welcome_message = {}
show_final_message = {}

loyalty_available_cache = None
loyalty_available_cache_time = datetime.min  # type: datetime
loyalty_available_lock = Lock()
five_seconds_timeout = 5000000


class LoyaltyInteractor(object):
    def __init__(
        self,
        pos_id=None,    # type: Optional[str]
    ):
        # type: (...) -> None

        self.pos_id = pos_id
        self.logger = logging.getLogger("PosActions")
        self.loyalty_id_type = None
        self.welcome_message_types = {
            "DOCUMENT": "$INSERT_CUSTOMER_DOC",
            "PHONE": "$INSERT_CUSTOMER_PHONE",
            "OTHER": "$INSERT_CUSTOMER_NUMBER",
        }

    def is_loyalty_available(
        self
    ):
        # type: (...) -> Optional[bool]

        global loyalty_available_cache_time
        global loyalty_available_cache

        with loyalty_available_lock:
            now = datetime.now()
            seconds_to_renew = 300
            renew_cache = now > loyalty_available_cache_time + timedelta(seconds=seconds_to_renew)
            if loyalty_available_cache is None or renew_cache:
                loyalty_available_cache = self._get_loyalty_status()
                loyalty_available_cache_time = now
                if loyalty_available_cache is None:
                    return None

            return loyalty_available_cache

    @staticmethod
    def _get_loyalty_status():
        # type: (...) -> Optional[bool]

        try:
            fifteen_seconds_timeout = 15000000
            response = send_message(
                dest_name="Loyalty",
                token=TK_LOYALTY_IS_AVAILABLE,
                format=FM_PARAM,
                timeout=fifteen_seconds_timeout,
            )
            return response.token == TK_SYS_ACK
        except (Exception,):
            return None

    def get_loyalty_id_type(
        self,
    ):
        # type: (...) -> str

        if self.loyalty_id_type:
            return self.loyalty_id_type

        response = send_message(
            dest_name="Loyalty",
            token=TK_LOYALTY_GET_ID_TYPE,
            format=FM_PARAM,
            data="LoyaltyIdentifierType",
            timeout=five_seconds_timeout,
        )
        if response.token == TK_SYS_ACK:
            self.loyalty_id_type = response.data

        return self.loyalty_id_type

    @staticmethod
    def set_loyalty_info(
        pos_id,         # type: int
        loyalty_id,     # type: str
        order_id,       # type: int
        table_id,       # type: str
        seat,           # type: str
    ):
        # type: (...) -> Optional[str]

        if not loyalty_id:
            return

        try:
            request = {
                "posId": pos_id,
                "loyaltyId": loyalty_id,
                "orderId": order_id,
                "tableId": table_id,
                "seat": seat,
            }
            ret = send_message(
                dest_name="Loyalty",
                token=TK_LOYALTY_GET_LOYALTY_ID,
                format=FM_PARAM,
                data=json.dumps(request).encode("utf-8"),
                timeout=five_seconds_timeout,
            )
            if ret.token == TK_SYS_ACK:
                return ret.data
            if ret.token == TK_LOYALTY_ID_NOT_FOUND:
                return None

            raise LoyaltyInteractor.LoyaltyPartnerError(ret.data)
        except MBException as ex:
            if "SE_NOTFOUND" in ex.message:
                return None

            raise ex

    def get_welcome_message_button_loyalty(
        self,
    ):
        # type: (...) -> (str, str)

        welcome_message = self.welcome_message_types.get(self.get_loyalty_id_type())
        return welcome_message, '$INFORM_LOYALTY_ID'

    def set_show_customer_welcome_message(
        self,
        show,    # type: bool
    ):
        # type: (...) -> None

        global show_welcome_message

        if self.pos_id not in show_welcome_message:
            show_welcome_message[self.pos_id] = show

    def get_show_customer_welcome_message(
        self,
    ):
        # type: (...) -> bool

        global show_welcome_message

        if self.pos_id in show_welcome_message:
            return show_welcome_message[self.pos_id]

        return False

    def set_show_customer_final_message(
        self,
        show,    # type: bool
    ):
        # type: (...) -> None

        global show_final_message

        if self.pos_id not in show_final_message:
            show_final_message[self.pos_id] = show

    def show_loyalty_customer_final_message(
        self,
        order_picture=None,      # type: Optional[eTree.Element]
    ):
        # type: (...) -> None

        self.logger.debug("showing customer final message")

        try:
            loyalty_customer_info = get_custom_property_value(
                pos_id=self.pos_id,
                custom_property="LOYALTY_CUSTOMER_INFO",
                order_picture=order_picture,
            )
            loyalty_customer_data = json.loads(loyalty_customer_info or "{}")
        except Exception:  # noqa
            self.exception("error getting/parsing loyalty customer info")
            loyalty_customer_data = {}

        customer_info = loyalty_customer_data.get("customerInfo", {})
        additional_info = customer_info.get("additionalInfo", {})
        end_sale_message = additional_info.get("endSaleMessage", None)

        if end_sale_message and self._show_final_message():
            title = "$LOYALTY_FINAL_MESSAGE_CUSTOMER_HEADER"
            message, qr_data = parse_qr_tag(text=end_sale_message)
            if qr_data:
                show_qr_code(
                    pos_id=self.pos_id,
                    code_data=qr_data,
                    buttons="$OK",
                    title=title,
                    info_text=message,
                )
            else:
                show_messagebox(
                    posid=self.pos_id,
                    message=message,
                    title=title,
                )

    def _show_final_message(
        self,
    ):
        # type: (...) -> bool

        global show_final_message

        if self.pos_id in show_final_message:
            return show_final_message[self.pos_id]

        return False

    class LoyaltyPartnerError(Exception):
        pass



# -*- coding: utf-8 -*-
import json

import requests
import sqlite3

from datetime import datetime
from requests import Response
from typing import Dict

from application.model import (
    UpdateData,
    UpdateStatus,
    UpdateType,
)
from application.model.configuration import Configurations
from application.model.customexception import (
    ErrorInsertingNewUpdateVersion,
    ErrorUpdatingStep,
)
from mbcontextmessagehandler import MbContextMessageBus
from mwhelper import (
    BaseRepository,
)
from persistence import Connection

from ._CommonPOSRepository import CommonPOSRepository


class BinaryUpdaterPOSRepository(BaseRepository):
    def __init__(self, message_bus, configs):
        # type: (MbContextMessageBus, Configurations) -> None

        super(BinaryUpdaterPOSRepository, self).__init__(message_bus.mbcontext)

        self.message_bus = message_bus
        self.configs = configs
        self.update_type = UpdateType.binaries
        self.update_value = self.update_type.value
        self.local_configs = configs.updaters[self.update_type.name]
        self.common_pos_repository = CommonPOSRepository(message_bus)

    def insert_new_binary_update(self, update_id, update_name):
        # type: (str, str) -> None

        try:
            def inner_func(conn):
                # type: (Connection) -> None

                date_now = str(datetime.now().isoformat())
                query = """
                INSERT OR REPLACE INTO UpdatesController
                (UpdateId, UpdateName, TypeId, StatusId, ObtainedDate)
                VALUES ('{0}', '{1}', '{2}', '{3}', '{4}')"""
                query_format = query.format(update_id, update_name, self.update_value, UpdateStatus.pending, date_now)
                conn.query(query_format)

            self.execute_with_connection(inner_func, service_name=self.configs.persistence_name)
        except Exception as ex:
            raise ErrorInsertingNewUpdateVersion(ex)

    def set_downloaded_date(self, update_id):
        self._update_step(update_id, 'DownloadedDate')

    def set_backup_date(self, update_id):
        self._update_step(update_id, 'BackupDate')

    def set_update_as_applied(
        self,                       # type: BinaryUpdaterPOSRepository
        update_id,                  # type: str
        backoffice_version,        # type: str
    ):
        # type: (...) -> None

        try:
            def inner_func(conn):
                # type: (Connection) -> None

                datetime_now = str(datetime.now().isoformat())
                query = """
                    UPDATE UpdatesController
                    SET
                        AppliedDate = '{0}',
                        ServerUpdateIdentifier = '{1}'
                    WHERE
                        Id = '{2}'
                        AND TypeId = '{3}'
                """.format(datetime_now, json.dumps(backoffice_version), update_id, self.update_value)
                conn.query(query)

            self.execute_with_connection(inner_func, service_name=self.configs.persistence_name)
        except Exception as ex:
            raise ErrorUpdatingStep(ex)

    def set_notified_date(
        self,           # type: BinaryUpdaterPOSRepository
        update_id,      # type: str
    ):
        # type: (...) -> None

        self._update_step(
            update_id=update_id,
            update_step='NotifiedDate',
        )

    def _update_step(
        self,               # type: BinaryUpdaterPOSRepository
        update_id,          # type: int
        update_step,        # type: str
    ):
        # type: (...) -> None

        try:
            def inner_func(conn):
                # type: (Connection) -> None

                datetime_now = str(datetime.now().isoformat())
                query = """
                    UPDATE UpdatesController
                    SET {0} = '{1}'
                    WHERE
                        Id = '{2}' AND
                        TypeId = '{3}'
                """.format(update_step, datetime_now, update_id, self.update_value)
                conn.query(query)

            self.execute_with_connection(inner_func, service_name=self.configs.persistence_name)
        except Exception as ex:
            raise ErrorUpdatingStep(ex)

    def get_last_binary_update_id(self):
        # type: () -> int

        def inner_func(conn):
            # type: (Connection) -> int

            query = """
                SELECT UpdateId
                FROM UpdatesController
                WHERE
                    TypeId = '{}' AND
                    AppliedDate IS NOT NULL
                ORDER BY Id DESC
                LIMIT 1""".format(self.update_value)

            cursor = conn.select(query)
            row = cursor.get_row(0)

            if row:
                update_id = row.get_entry("UpdateId")
                return int(update_id)

            return 0

        return self.execute_with_connection(inner_func, service_name=self.configs.persistence_name)

    def get_last_binary_update_info(self):
        # type: () -> UpdateData

        def inner_func(conn):
            # type: (Connection) -> UpdateData

            query = """
                SELECT
                    Id,
                    UpdateName,
                    AppliedDate,
                    NotifiedDate,
                    ServerUpdateIdentifier
                FROM UpdatesController
                WHERE
                    TypeId = '{}' AND
                    ObtainedDate IS NOT NULL
                ORDER BY Id DESC
                LIMIT 1
            """.format(self.update_value)

            cursor = conn.select(query)
            row = cursor.get_row(0)

            if row:
                return UpdateData(
                    update_id=int(row.get_entry("Id")),
                    update_name=row.get_entry("UpdateName"),
                    applied_date=row.get_entry("AppliedDate"),
                    notified_date=row.get_entry("NotifiedDate"),
                    server_update_identifier=row.get_entry("ServerUpdateIdentifier"),
                )

            return UpdateData(
                update_id=0,
            )

        return self.execute_with_connection(inner_func, service_name=self.configs.persistence_name)

    def get_last_binary_update_not_applied_controller_id(self):
        # type: () -> Dict

        def inner_func(conn):
            # type: (Connection) -> Dict

            query = """
                SELECT Id, DownloadedDate, UpdateName
                FROM UpdatesController
                WHERE TypeId = '{}'
                AND ObtainedDate IS NOT NULL
                AND DownloadedDate IS NOT NULL
                AND AppliedDate IS NULL
                ORDER BY Id DESC
                LIMIT 1""".format(self.update_value)

            cursor = conn.select(query)
            row = cursor.get_row(0)

            if row:
                update_id = row.get_entry("Id")
                downloaded_date = row.get_entry("DownloadedDate")
                update_name = row.get_entry("UpdateName")
                result = {
                    "update_id": update_id,
                    "downloaded_date": datetime.strptime(downloaded_date, "%Y-%m-%dT%H:%M:%S.%f"),
                    "update_name": update_name
                }
                return result

            return {}

        return self.execute_with_connection(inner_func, service_name=self.configs.persistence_name)

    def get_last_binary_update_name(self):
        # type: () -> str

        def inner_func(conn):
            # type: (Connection) -> str

            query = """
                SELECT UpdateName
                FROM UpdatesController
                WHERE TypeId = '{}' AND AppliedDate IS NOT NULL
                ORDER BY Id DESC
                LIMIT 1""".format(self.update_value)

            cursor = conn.select(query)
            row = cursor.get_row(0)

            if row:
                update_name = row.get_entry("UpdateName")
                return update_name

            return ""

        return self.execute_with_connection(inner_func, service_name=self.configs.persistence_name)

    def invalidate_last_binary_update(
        self,       # type: BinaryUpdaterPOSRepository
    ):
        # type: (...) -> None

        try:
            def inner_func(conn):
                # type: (Connection) -> None

                query = """
                    UPDATE UpdatesController
                    SET AppliedDate = NULL, NotifiedDate = NULL
                    WHERE
                        Id = (
                            SELECT Max(Id)
                            FROM UpdatesController
                            WHERE
                                TypeId = '{}'
                            ORDER BY Id DESC
                        )
                """.format(self.update_value)
                conn.query(query)

            self.execute_with_connection(inner_func, service_name=self.configs.persistence_name)
        except Exception as ex:
            raise ErrorUpdatingStep(ex)

    def remove_downloaded_update(self, update_id):
        # type: (int) -> None
        try:
            def inner_func(conn):
                # type: (Connection) -> None

                query = """
                            DELETE FROM UpdatesController
                            WHERE Id = '{0}' AND TypeId = '{1}'
                        """.format(update_id, self.update_value)
                conn.query(query)

            self.execute_with_connection(inner_func, service_name=self.configs.persistence_name)
        except Exception as ex:
            raise ErrorUpdatingStep(ex)

    def has_already_applied_update(self):
        # type: () -> bool

        return self.common_pos_repository.has_already_applied_update(self.configs.persistence_name, self.update_value)

    @staticmethod
    def download_from_url(
        url     # type: str
    ):
        # type: (...) -> Response

        return requests.get(url, allow_redirects=True)

    def download_chrome_and_save(
        self,
        download_url,   # type: str
        file_path       # type: str
    ):
        # type: (...) -> bool

        response = self.download_from_url(url=download_url)
        if response.status_code == 200:
            with open(file_path, "wb") as zip_object:
                for chunk in response.iter_content(chunk_size=128):
                    zip_object.write(chunk)
            return True
        return False

    @staticmethod
    def _create_connection(db_file):
        return sqlite3.connect(db_file)



class CannotSaveOrderWithNotFilledMandatoryOption(Exception):

    def __init__(
        self,       # type: CannotSaveOrderWithNotFilledMandatoryOption
        part_code,  # type: int
        part_name,  # type: str
    ):
        # type: (...) -> None
        super(CannotSaveOrderWithNotFilledMandatoryOption, self).__init__()

        self.part_code = part_code
        self.part_name = part_name

    @staticmethod
    def get_error_code():
        # type: (...) -> str

        return "CannotSaveOrderWithNotFilledMandatoryOption"

    def __str__(
        self,   # type: CannotSaveOrderWithNotFilledMandatoryOption
    ):
        # type: (...) -> str

        return "{}|{}|{}".format(self.get_error_code(), self.part_code, self.part_name)
